"""
Morphology operations on multi-label ANTsImage types
"""

__all__ = ['multi_label_morphology']

import numpy as np
from ants.decorators import image_method

@image_method
def multi_label_morphology(image, operation, radius, dilation_mask=None, label_list=None, force=False):
    """
    Morphology on multi label images.
    
    Wraps calls to iMath binary morphology. Additionally, dilation and closing operations preserve
    pre-existing labels. The choices of operation are:
    
    Dilation: dilates all labels sequentially, but does not overwrite original labels.
    This reduces dependence on the intensity ordering of adjoining labels. Ordering dependence
    can still arise if two or more labels dilate into the same space - in this case, the label
    with the lowest intensity is retained. With a mask, dilated labels are multiplied by the
    mask and then added to the original label, thus restricting dilation to the mask region.
    
    Erosion: Erodes labels independently, equivalent to calling iMath iteratively.
    
    Closing: Close holes in each label sequentially, but does not overwrite original labels.
    
    Opening: Opens each label independently, equivalent to calling iMath iteratively.
    
    Arguments
    ---------
    image : ANTsImage
        Input image should contain only 0 for background and positive integers for labels.
    
    operation : string
        One of MD, ME, MC, MO, passed to iMath.
    
    radius : integer
        radius of the morphological operation.
    
    dilation_mask : ANTsImage
        Optional binary mask to constrain dilation only (eg dilate cortical label into WM).
    
    label_list : list or tuple or numpy.ndarray
        Optional list of labels, to perform operation upon. Defaults to all unique 
        intensities in image.

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_data('r16'))
    >>> labels = ants.get_mask(img,1,150) + ants.get_mask(img,151,225) * 2
    >>> labels_dilated = ants.multi_label_morphology(labels, 'MD', 2)
    >>> # should see original label regions preserved in dilated version
    >>> # label N should have mean N and 0 variance
    >>> print(ants.label_stats(labels_dilated, labels))
    """
    if (label_list is None) or (len(label_list) == 1):
        label_list = np.sort(np.unique(image[image > 0]))
        if (len(label_list) > 200) and (not force):
            raise ValueError('More than 200 labels... Make sure the image is discrete'
                             ' and call this function again with `force=True` if you'
                             ' really want to do this.')

    image_binary = image.clone()
    image_binary[image_binary > 1] = 1

    # Erosion / opening is simply a case of looping over the input labels
    if (operation == 'ME') or (operation == 'MO'):
        output = image.clone()

        for current_label in label_list:
            output = output.iMath(operation, radius, current_label)

        return output

    if dilation_mask is not None:
        if int(dilation_mask.max()) != 1:
            raise ValueError('Mask is either empty or not binary')

    output = image.clone()
    for current_label in label_list:
        current_label_region = image.threshold_image(current_label, current_label)
        other_labels = output - current_label_region
        clab_binary_morphed = current_label_region.iMath(operation, radius, 1)

        if (operation == 'MD') and (dilation_mask is not None):
            clab_binary_morphed_nooverlap = current_label_region + dilation_mask * clab_binary_morphed - other_labels
            clab_binary_morphed_nooverlap = clab_binary_morphed_nooverlap.threshold_image(1, 2)
        else:
            clab_binary_morphed_nooverlap = clab_binary_morphed - other_labels
            clab_binary_morphed_nooverlap = clab_binary_morphed_nooverlap.threshold_image(1, 1)

        output = output + clab_binary_morphed_nooverlap * current_label

    return output




__all__ = ["segmentation_to_one_hot,"
           "one_hot_to_segmentation"]

import ants
import numpy as np

def segmentation_to_one_hot(segmentations_array,
                            segmentation_labels=None,
                            channel_first_ordering=False):

    """
    Basic one-hot transformation of segmentations array

    Arguments
    ---------
    segmentations_array : numpy array
        multi-label numpy array

    segmentation_labels : tuple or list
        Note that a background label (typically 0) needs to be included.
        
    channel_first_ordering : bool
        Specifies the ordering of the dimensions.

    Returns
    -------
    A numpy array where the segmentation labels are expanded in a one-hot
    fashion in the channels dimension.

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> seg = ants.kmeans_segmentation(image, 3)['segmentation']
    >>> one_hot = segmentation_to_one_hot(seg.numpy().astype('int'))
    """

    if segmentation_labels is None:
        segmentation_labels = np.unique(segmentations_array)

    number_of_labels = len(segmentation_labels)
    if number_of_labels < 2:
        raise ValueError("At least two segmentation labels need to be specified.")

    image_dimension = len(segmentations_array.shape)

    one_hot_array = np.zeros((*segmentations_array.shape, number_of_labels))
    for i in range(number_of_labels):
        per_label = np.zeros_like(segmentations_array)
        per_label[segmentations_array == segmentation_labels[i]] = 1
        if image_dimension == 2:
            one_hot_array[:,:,i] = per_label
        elif image_dimension == 3:
            one_hot_array[:,:,:,i] = per_label
        else:
            raise ValueError("Unrecognized image dimensionality.")

    if channel_first_ordering:  
        if image_dimension == 2:
            one_hot_array = one_hot_array.transpose((2, 0, 1))
        elif image_dimension == 3:
            one_hot_array = one_hot_array.transpose((3, 0, 1, 2))

    return one_hot_array


def one_hot_to_segmentation(one_hot_array,
                            domain_image,
                            channel_first_ordering=False):

    """
    Inverse of basic one-hot transformation of segmentations array

    Arguments
    ---------
    one_hot_array : an array
        Numpy array where the channel dimension contains the one-hot
        encoding.

    domain_image : ANTs image
        Defines the geometry of the returned probability images

    channel_first_ordering : bool
        Specifies the ordering of the dimensions.
        
    Returns
    -------
    List of probability images.

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    """

    number_of_labels = one_hot_array.shape[-1]
    if channel_first_ordering:
        number_of_labels = one_hot_array.shape[0]

    image_dimension = domain_image.dimension

    probability_images = list()
    for label in range(number_of_labels):
        if image_dimension == 2:
            if channel_first_ordering:
                image_array = np.squeeze(one_hot_array[label,:,:])
            else:         
                image_array = np.squeeze(one_hot_array[:,:,label])
        else:
            if channel_first_ordering:
                image_array = np.squeeze(one_hot_array[label,:,:,:])
            else:         
                image_array = np.squeeze(one_hot_array[:,:,:,label])
        probability_images.append(
            ants.from_numpy_like(image_array, domain_image))

    return probability_images





def image_type_cast(image_list, pixeltype=None):
    """
    Cast a list of images to the highest pixeltype present in the list
    or all to a specified type

    ANTsR function: `antsImageTypeCast`

    Arguments
    ---------
    image_list : list/tuple
        images to cast

    pixeltype : string (optional)
        pixeltype to cast to. If None, images will be cast to the highest
        precision pixeltype found in image_list

    Returns
    -------
    list of ANTsImages
        given images casted to new type
    """
    if not isinstance(image_list, (list,tuple)):
        raise ValueError('image_list must be list of ANTsImage types')

    pixtypes = []
    for img in image_list:
        pixtypes.append(img.pixeltype)

    if pixeltype is None:
        pixeltype = 'unsigned char'
        for p in pixtypes:
            if p == 'double':
                pixeltype = 'double'
            elif (p=='float') and (pixeltype!='double'):
                pixeltype = 'float'
            elif (p=='unsigned int') and (pixeltype!='float') and (pixeltype!='double'):
                pixeltype = 'unsigned int'

    out_images = []
    for img in image_list:
        if img.pixeltype == pixeltype:
            out_images.append(img)
        else:
            out_images.append(img.clone(pixeltype))

    return out_images


 

__all__ = ['reflect_image']

from tempfile import mktemp

import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn

@image_method
def reflect_image(image, axis=None, tx=None, metric='mattes'):
    """
    Reflect an image along an axis

    ANTsR function: `reflectImage`

    Arguments
    ---------
    image : ANTsImage
        image to reflect
    
    axis : integer (optional)
        which dimension to reflect across, numbered from 0 to imageDimension-1
    
    tx : string (optional)
        transformation type to estimate after reflection
    
    metric : string  
        similarity metric for image registration. see antsRegistration.
    
    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data('r16'), 'float' )
    >>> axis = 2
    >>> asym = ants.reflect_image(fi, axis, 'Affine')['warpedmovout']
    >>> asym = asym - fi
    """
    if axis is None:
        axis = image.dimension - 1

    if (axis > image.dimension) or (axis < 0):
        axis = image.dimension - 1

    rflct = mktemp(suffix='.mat')

    libfn = get_lib_fn('reflectionMatrix')
    libfn(image.pointer, axis, rflct)

    if tx is not None:
        rfi = ants.registration(image, image, type_of_transform=tx,
                            syn_metric=metric, outprefix=mktemp(),
                            initial_transform=rflct)
        return rfi
    else:
        return ants.apply_transforms(image, image, rflct)





from .plot import plot
from .movie import movie
from .plot_hist import plot_hist
from .plot_grid import plot_grid
from .plot_ortho import plot_ortho
from .plot_ortho_stack import plot_ortho_stack
from .plot_directory import plot_directory



"""
Joint Label Fusion algorithm
"""

__all__ = ["joint_label_fusion", "local_joint_label_fusion"]

import os
import numpy as np
import warnings
from pathlib import Path
from tempfile import NamedTemporaryFile
from tempfile import mktemp
import glob
import re
import math

import ants
from ants.internal import get_lib_fn, get_pointer_string, process_arguments


def joint_label_fusion(
    target_image,
    target_image_mask,
    atlas_list,
    beta=4,
    rad=2,
    label_list=None,
    rho=0.01,
    usecor=False,
    r_search=3,
    nonnegative=False,
    no_zeroes=False,
    max_lab_plus_one=False,
    output_prefix=None,
    verbose=False,
):
    """
    A multiple atlas voting scheme to customize labels for a new subject.
    This function will also perform intensity fusion. It almost directly
    calls the C++ in the ANTs executable so is much faster than other
    variants in ANTsR.

    One may want to normalize image intensities for each input image before
    passing to this function. If no labels are passed, we do intensity fusion.
    Note on computation time: the underlying C++ is multithreaded.
    You can control the number of threads by setting the environment
    variable ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS e.g. to use all or some
    of your CPUs. This will improve performance substantially.
    For instance, on a macbook pro from 2015, 8 cores improves speed by about 4x.

    ANTsR function: `jointLabelFusion`

    Arguments
    ---------
    target_image : ANTsImage
        image to be approximated

    target_image_mask : ANTsImage
        mask with value 1

    atlas_list : list of ANTsImage types
        list containing intensity images

    beta : scalar
        weight sharpness, default to 2

    rad : scalar
        neighborhood radius, default to 2

    label_list : list of ANTsImage types (optional)
        list containing images with segmentation labels

    rho : scalar
        ridge penalty increases robustness to outliers but also makes image converge to average

    usecor : boolean
        employ correlation as local similarity

    r_search : scalar
        radius of search, default is 3

    nonnegative : boolean
        constrain weights to be non-negative

    no_zeroes : boolean
        this will constrain the solution only to voxels that are always non-zero in the label list

    max_lab_plus_one : boolean
        this will add max label plus one to the non-zero parts of each label where the target mask
        is greater than one.  NOTE: this will have a side effect of adding to the original label
        images that are passed to the program.  It also guarantees that every position in the
        labels have some label, rather than none.  Ie it guarantees to explicitly parcellate the
        input data.

    output_prefix: string
        file prefix for storing output probabilityimages to disk

    verbose : boolean
        whether to show status updates

    Returns
    -------
    dictionary w/ following key/value pairs:
        `segmentation` : ANTsImage
            segmentation image

        `intensity` : ANTsImage
            intensity image

        `probabilityimages` : list of ANTsImage types
            probability map image for each label

        `segmentation_numbers` : list of numbers
            segmentation label (number, int) for each probability map


    Example
    -------
    >>> import ants
    >>> ref = ants.image_read( ants.get_ants_data('r16'))
    >>> ref = ants.resample_image(ref, (50,50),1,0)
    >>> ref = ants.iMath(ref,'Normalize')
    >>> mi = ants.image_read( ants.get_ants_data('r27'))
    >>> mi2 = ants.image_read( ants.get_ants_data('r30'))
    >>> mi3 = ants.image_read( ants.get_ants_data('r62'))
    >>> mi4 = ants.image_read( ants.get_ants_data('r64'))
    >>> mi5 = ants.image_read( ants.get_ants_data('r85'))
    >>> refmask = ants.get_mask(ref)
    >>> refmask = ants.iMath(refmask,'ME',2) # just to speed things up
    >>> ilist = [mi,mi2,mi3,mi4,mi5]
    >>> seglist = [None]*len(ilist)
    >>> for i in range(len(ilist)):
    >>>     ilist[i] = ants.iMath(ilist[i],'Normalize')
    >>>     mytx = ants.registration(fixed=ref , moving=ilist[i] ,
    >>>         type_of_transform = ('Affine') )
    >>>     mywarpedimage = ants.apply_transforms(fixed=ref,moving=ilist[i],
    >>>             transformlist=mytx['fwdtransforms'])
    >>>     ilist[i] = mywarpedimage
    >>>     seg = ants.threshold_image(ilist[i],'Otsu', 3)
    >>>     seglist[i] = ( seg ) + ants.threshold_image( seg, 1, 3 ).morphology( operation='dilate', radius=3 )
    >>> r = 2
    >>> pp = ants.joint_label_fusion(ref, refmask, ilist, r_search=2,
    >>>                     label_list=seglist, rad=[r]*ref.dimension )
    >>> pp = ants.joint_label_fusion(ref,refmask,ilist, r_search=2, rad=[r]*ref.dimension)
    """
    segpixtype = "unsigned int"
    if (label_list is None) or (np.any([l is None for l in label_list])):
        doJif = True
    else:
        doJif = False

    if not doJif:
        if len(label_list) != len(atlas_list):
            raise ValueError("len(label_list) != len(atlas_list)")
        if no_zeroes:
            for label in label_list:
                target_image_mask[label == 0] = 0
        inlabs = set()
        for label in label_list:
            values = np.unique(label[target_image_mask != 0 and label != 0])
            inlabs = inlabs.union(values)
        inlabs = sorted(inlabs)
        maxLab = max(inlabs)
        if max_lab_plus_one:
            for label in label_list:
                label[label == 0] = maxLab + 1
        mymask = target_image_mask.clone()
    else:
        mymask = target_image_mask

###### security issues with mktemp but could not figure out the right solution
###### NamedTemporaryFile creates a file with permissions:
###### -rw-------  1 stnava  staff
###### whereas mktemp gives
###### -rw-r--r--  1 stnava  staff
###### the latter is what we want - one solution is to use chmod via os but
###### am currently too lazy to change one line of code to two or more everywhere

#    osegfn = NamedTemporaryFile(prefix="antsr", suffix="myseg.nii.gz",delete=False).name
    osegfn = mktemp(prefix="antsr", suffix="myseg.nii.gz")
    # segdir = osegfn.replace(os.path.basename(osegfn),'')

    if os.path.exists(osegfn):
        os.remove(osegfn)

    if output_prefix is None:
#        probs = NamedTemporaryFile(prefix="antsr", suffix="prob%02d.nii.gz",delete=False).name
        probs = mktemp(prefix="antsr", suffix="prob%02d.nii.gz")
        probsbase = os.path.basename(probs)
        tdir = probs.replace(probsbase, "")
        searchpattern = probsbase.replace("%02d", "*")

    if output_prefix is not None:
        probs = output_prefix + "prob%02d.nii.gz"
        probpath = Path(probs).parent
        Path(probpath).mkdir(parents=True, exist_ok=True)
        probsbase = os.path.basename(probs)
        tdir = probs.replace(probsbase, "")
        searchpattern = probsbase.replace("%02d", "*")

    mydim = target_image_mask.dimension
    if not doJif:
        # not sure if these should be allocated or what their size should be
        outimg = label_list[1].clone(segpixtype)
        outimgi = target_image * 0

        outimg_ptr = get_pointer_string(outimg)
        outimgi_ptr = get_pointer_string(outimgi)
        outs = "[%s,%s,%s]" % (outimg_ptr, outimgi_ptr, probs)
    else:
        outimgi = target_image * 0
        outs = get_pointer_string(outimgi)

    mymask = mymask.clone(segpixtype)
    if (not isinstance(rad, (tuple, list))) or (len(rad) == 1):
        myrad = [rad] * mydim
    else:
        myrad = rad

    if len(myrad) != mydim:
        raise ValueError("path radius dimensionality must equal image dimensionality")

    myrad = "x".join([str(mr) for mr in myrad])
    vnum = 1 if verbose else 0
    nnum = 1 if nonnegative else 0
    mypc = "MSQ"
    if usecor:
        mypc = "PC"

    myargs = {
        "d": mydim,
        "t": target_image,
        "a": rho,
        "b": beta,
        "c": nnum,
        "p": myrad,
        "m": mypc,
        "s": r_search,
        "x": mymask,
        "o": outs,
        "v": vnum,
    }

    kct = len(myargs.keys())
    for k in range(len(atlas_list)):
        kct += 1
        myargs["g-MULTINAME-%i" % kct] = atlas_list[k]
        if not doJif:
            kct += 1
            castseg = label_list[k].clone(segpixtype)
            myargs["l-MULTINAME-%i" % kct] = castseg

    myprocessedargs = process_arguments(myargs)

    libfn = get_lib_fn("antsJointFusion")
    rval = libfn(myprocessedargs)
    if rval != 0:
        print("Warning: Non-zero return from antsJointFusion")

    if doJif:
        return outimgi

    probsout = glob.glob(os.path.join(tdir, "*" + searchpattern))
    probsout.sort()
    probimgs = []
#    print( os.system("ls -l "+probsout[0]) )
    for idx in range(len(probsout)):
        probimgs.append(ants.image_read(probsout[idx]))

    #    if len(probsout) != (len(inlabs)) and max_lab_plus_one == False:
    #        warnings.warn("Length of output probabilities != length of unique input labels")

    segmentation_numbers = [0] * len(probsout)
    for i in range(len(probsout)):
        temp = str.split(probsout[i], "prob")
        segnum = temp[len(temp) - 1].split(".nii.gz")[0]
        segmentation_numbers[i] = int(segnum)

    if max_lab_plus_one == False:
        segmat = ants.images_to_matrix(probimgs, target_image_mask)
        finalsegvec = segmat.argmax(axis=0)
        finalsegvec2 = finalsegvec.copy()
        # mapfinalsegvec to original labels
        for i in range(len(probsout)):
            temp = str.split(probsout[i], "prob")
            segnum = temp[len(temp) - 1].split(".nii.gz")[0]
            finalsegvec2[finalsegvec == i] = segnum
        outimg = ants.make_image(target_image_mask, finalsegvec2)

        return {
            "segmentation": outimg,
            "intensity": outimgi,
            "probabilityimages": probimgs,
            "segmentation_numbers": segmentation_numbers,
        }

    if max_lab_plus_one == True:
        mymaxlab = max(segmentation_numbers)
        matchings_indices = [
            i
            for i, segmentation_numbers in enumerate(segmentation_numbers)
            if segmentation_numbers == mymaxlab
        ]
        background_prob = probimgs[matchings_indices[0]]
        background_probfn = probsout[matchings_indices[0]]
        del probimgs[matchings_indices[0]]
        del probsout[matchings_indices[0]]
        del segmentation_numbers[matchings_indices[0]]

        segmat = ants.images_to_matrix(probimgs, target_image_mask)

        finalsegvec = segmat.argmax(axis=0)
        finalsegvec2 = finalsegvec.copy()
        # mapfinalsegvec to original labels
        for i in range(len(probsout)):
            temp = str.split(probsout[i], "prob")
            segnum = temp[len(temp) - 1].split(".nii.gz")[0]
            finalsegvec2[finalsegvec == i] = segnum

        outimg = ants.make_image(target_image_mask, finalsegvec2)

        # next decide what is "background" based on the sum of the first k labels vs the prob of the last one
        firstK = probimgs[0] * 0
        for i in range(len(probsout)):
            firstK = firstK + probimgs[i]

        segmat = ants.images_to_matrix([background_prob, firstK], target_image_mask)
        bkgsegvec = segmat.argmax(axis=0)
        outimg = outimg * ants.make_image(target_image_mask, bkgsegvec)

        return {
            "segmentation": outimg * ants.make_image(target_image_mask, bkgsegvec),
            "segmentation_raw": outimg,
            "intensity": outimgi,
            "probabilityimages": probimgs,
            "segmentation_numbers": segmentation_numbers,
            "background_prob": background_prob,
        }


def local_joint_label_fusion(
    target_image,
    which_labels,
    target_mask,
    initial_label,
    atlas_list,
    label_list,
    submask_dilation=10,
    type_of_transform="SyN",
    aff_metric="meansquares",
    syn_metric="mattes",
    syn_sampling=32,
    reg_iterations=(40, 20, 0),
    aff_iterations=(500, 50, 0),
    grad_step=0.2,
    flow_sigma=3,
    total_sigma=0,
    beta=4,
    rad=2,
    rho=0.1,
    usecor=False,
    r_search=3,
    nonnegative=False,
    no_zeroes=False,
    max_lab_plus_one=False,
    local_mask_transform="Similarity",
    output_prefix=None,
    verbose=False,
):
    """
    A local version of joint label fusion that focuses on a subset of labels.
    This is primarily different from standard JLF because it performs
    registration on the label subset and focuses JLF on those labels alone.

    ANTsR function: `localJointLabelFusion`

    Arguments
    ---------
    target_image : ANTsImage
        image to be labeled

    which_labels : numeric vector
        label number(s) that exist(s) in both the template and library

    target_image_mask : ANTsImage
        a mask for the target image (optional), passed to joint fusion

    initial_label : ANTsImage
        initial label set, may be same labels as library or binary.
        typically labels would be produced by a single deformable registration
        or by manual labeling.

    atlas_list : list of ANTsImage types
        list containing intensity images

    label_list : list of ANTsImage types (optional)
        list containing images with segmentation labels

    submask_dilation : integer
        amount to dilate initial mask to define region on which
        we perform focused registration

    type_of_transform : string
        A linear or non-linear registration type. Mutual information metric by default.
        See Notes below for more.

    aff_metric : string
        the metric for the affine part (GC, mattes, meansquares)

    syn_metric : string
        the metric for the syn part (CC, mattes, meansquares, demons)

    syn_sampling : scalar
        the nbins or radius parameter for the syn metric

    reg_iterations : list/tuple of integers
        vector of iterations for syn. we will set the smoothing and multi-resolution parameters based on the length of this vector.


    aff_iterations : list/tuple of integers
        vector of iterations for low-dimensional registration.

    grad_step : scalar
        gradient step size (not for all tx)

    flow_sigma : scalar
        smoothing for update field

    total_sigma : scalar
        smoothing for total field

    beta : scalar
        weight sharpness, default to 2

    rad : scalar
        neighborhood radius, default to 2

    rho : scalar
        ridge penalty increases robustness to outliers but also makes image converge to average

    usecor : boolean
        employ correlation as local similarity

    r_search : scalar
        radius of search, default is 3

    nonnegative : boolean
        constrain weights to be non-negative

    no_zeroes : boolean
        this will constrain the solution only to voxels that are always non-zero in the label list

    max_lab_plus_one : boolean
        this will add max label plus one to the non-zero parts of each label where the target mask
        is greater than one.  NOTE: this will have a side effect of adding to the original label
        images that are passed to the program.  It also guarantees that every position in the
        labels have some label, rather than none.  Ie it guarantees to explicitly parcellate the
        input data.

    local_mask_transform: string
        the type of transform for the local mask alignment - usually translation,
        rigid, similarity or affine.

    output_prefix: string
        file prefix for storing output probabilityimages to disk

    verbose : boolean
        whether to show status updates

    Returns
    -------
    dictionary w/ following key/value pairs:
        `segmentation` : ANTsImage
            segmentation image

        `intensity` : ANTsImage
            intensity image

        `probabilityimages` : list of ANTsImage types
            probability map image for each label

    """
    myregion = ants.mask_image(initial_label, initial_label, which_labels)
    if myregion.max() == 0:
        myregion = ants.threshold_image(initial_label, 1, math.inf)

    myregionb = ants.threshold_image(myregion, 1, math.inf)
    myregionAroundRegion = ants.iMath(myregionb, "MD", submask_dilation)
    if target_mask is not None:
        myregionAroundRegion = myregionAroundRegion * target_mask
    croppedImage = ants.crop_image(target_image, myregionAroundRegion)
    croppedMask = ants.crop_image(myregionAroundRegion, myregionAroundRegion)
    mycroppedregion = ants.crop_image(myregion, myregionAroundRegion)
    croppedmappedImages = []
    croppedmappedSegs = []
    if verbose is True:
        print("Begin registrations:")
    for k in range(len(atlas_list)):

        if verbose is True:
            print(str(k) + "...")

        if verbose is True:
            print( "local-seg-tx: " + local_mask_transform )
        libregion = ants.mask_image(label_list[k], label_list[k], which_labels)
        initMap = ants.registration(
            mycroppedregion, libregion, type_of_transform=local_mask_transform, aff_metric=aff_metric, aff_iterations=aff_iterations, verbose=False
        )["fwdtransforms"]
        if verbose is True:
            print( "local-img-tx: " + type_of_transform )
        localReg = ants.registration(
            croppedImage,
            atlas_list[k],
            reg_iterations=reg_iterations,
            flow_sigma=flow_sigma,
            total_sigma=total_sigma,
            grad_step=grad_step,
            type_of_transform=type_of_transform,
            syn_metric=syn_metric,
            syn_sampling=syn_sampling,
            initial_transform=initMap[0],
            verbose=False,
        )
        transformedImage = ants.apply_transforms(
            croppedImage, atlas_list[k], localReg["fwdtransforms"]
        )
        transformedLabels = ants.apply_transforms(
            croppedImage,
            label_list[k],
            localReg["fwdtransforms"],
            interpolator="nearestNeighbor",
        )
        croppedmappedImages.append(transformedImage)
        croppedmappedSegs.append(transformedLabels)

    ljlf = joint_label_fusion(
        croppedImage,
        croppedMask,
        atlas_list=croppedmappedImages,
        label_list=croppedmappedSegs,
        beta=beta,
        rad=rad,
        rho=rho,
        usecor=usecor,
        r_search=r_search,
        nonnegative=nonnegative,
        no_zeroes=no_zeroes,
        max_lab_plus_one=max_lab_plus_one,
        output_prefix=output_prefix,
        verbose=verbose,
    )

    return {
        "ljlf": ljlf,
        "croppedImage": croppedImage,
        "croppedmappedImages": croppedmappedImages,
        "croppedmappedSegs": croppedmappedSegs,
    }






__all__ = ['kmeans_segmentation']

import ants

def kmeans_segmentation(image, k, kmask=None, mrf=0.1):
    """
    K-means image segmentation that is a wrapper around `ants.atropos`

    ANTsR function: `kmeansSegmentation`

    Arguments
    ---------
    image : ANTsImage
        input image

    k : integer
        integer number of classes

    kmask : ANTsImage (optional)
        segment inside this mask

    mrf : scalar
        smoothness, higher is smoother

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('r16'), 'float')
    >>> fi = ants.n3_bias_field_correction(fi, 2)
    >>> seg = ants.kmeans_segmentation(fi, 3)
    """
    dim = image.dimension
    kmimage = ants.iMath(image, 'Normalize')
    if kmask is None:
        kmask = ants.get_mask(kmimage, 0.01, 1, cleanup=2)
    kmask = ants.iMath(kmask, 'FillHoles').threshold_image(1,2)
    nhood = 'x'.join(['1']*dim)
    mrf = '[%s,%s]' % (str(mrf), nhood)
    kmimage = ants.atropos(a = kmimage, m = mrf, c = '[5,0]', i = 'kmeans[%s]'%(str(k)), x = kmask)
    kmimage['segmentation'] = kmimage['segmentation'].clone(image.pixeltype)
    return kmimage




__all__ = ['invert_displacement_field']

import ants
from ants.internal import get_lib_fn


def invert_displacement_field(displacement_field, 
                              inverse_field_initial_estimate, 
                              maximum_number_of_iterations=20, 
                              mean_error_tolerance_threshold=0.001, 
                              max_error_tolerance_threshold=0.1,
                              enforce_boundary_condition=True):
    """
    Invert displacement field.

    Arguments
    ---------
    displacement_field : ANTsImage displacement field
        displacement field

    inverse_field_initial_estimate : ANTsImage displacement field
        initial guess

    maximum_number_of_iterations : integer
        number of iterations

    mean_error_tolerance_threshold : float
        mean error tolerance threshold

    max_error_tolerance_threshold : float
        max error tolerance threshold

    enforce_boundary_condition : bool
        enforce stationary boundary condition
        

    Example
    -------
    >>> import ants
    """

    libfn = get_lib_fn('invertDisplacementFieldD%i' % displacement_field.dimension)
    inverse_field = libfn(displacement_field.pointer, inverse_field_initial_estimate.pointer, 
        maximum_number_of_iterations, mean_error_tolerance_threshold, 
        max_error_tolerance_threshold, enforce_boundary_condition)

    new_image = ants.from_pointer(inverse_field).clone('float')
    return new_image






__all__ = ['mni2tal']

def mni2tal(xin):
    """
    mni2tal for converting from ch2/mni space to tal - very approximate.

    This is a standard approach but it's not very accurate.

    ANTsR function: `mni2tal`

    Arguments
    ---------
    xin : tuple
        point in mni152 space.

    Returns
    -------
    tuple

    Example
    -------
    >>> import ants
    >>> ants.mni2tal( (10,12,14) )

    References
    ----------
    http://bioimagesuite.yale.edu/mni2tal/501_95733_More\\%20Accurate\\%20Talairach\\%20Coordinates\\%20SLIDES.pdf
    http://imaging.mrc-cbu.cam.ac.uk/imaging/MniTalairach
    """
    if (not isinstance(xin, (tuple,list))) or (len(xin) != 3):
        raise ValueError('xin must be tuple/list with 3 coordinates')

    x = list(xin)
    # The input image is in RAS coordinates but we use ITK which returns LPS
    # coordinates.  So we need to flip the coordinates such that L => R and P => A to
    # get RAS (MNI) coordinates
    x[0] = x[0] * (-1)  # flip X
    x[1] = x[1] * (-1)  # flip Y
    
    xout = x
    
    if (x[2] >= 0):
        xout[0] = x[0] * 0.99
        xout[1] = x[1] * 0.9688 + 0.046 * x[2]
        xout[2] = x[1] * (-0.0485) + 0.9189 * x[2]
    
    if (x[2] < 0):
        xout[0] = x[0] * 0.99
        xout[1] = x[1] * 0.9688 + 0.042 * x[2]
        xout[2] = x[1] * (-0.0485) + 0.839 * x[2]
    
    return(xout)












__all__ = ['create_jacobian_determinant_image',
           'deformation_gradient']

from tempfile import NamedTemporaryFile

import ants
import numpy as np
from ants.internal import get_lib_fn, process_arguments


def deformation_gradient( warp_image, to_rotation=False, to_inverse_rotation=False, py_based=False ):
    """
    Compute the deformation gradient from an image containing a warp (deformation).

    This function now includes a highly optimized pure Python/NumPy implementation.

    ANTsR function: `NA`

    Arguments
    ---------
    warp_image : ANTsImage (or filename if not py_based)
        image that defines the deformation field (vector pixels)

    to_rotation : boolean
        maps deformation gradient to a rotation matrix using polar decomposition.

    to_inverse_rotation : boolean
        map the deformation gradient to a rotation matrix, and return its inverse.
        This is useful for reorienting tensors and vectors after resampling.

    py_based: boolean
        If True, uses the optimized pure Python/NumPy implementation. If False,
        uses the classic C++ backend.

    Returns
    -------
    ANTsImage with dimension*dimension components indexed in order U_xyz, V_xyz, W_xyz
        where U is the x-component of deformation and xyz are spatial.

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data('r16'))
    >>> mi = ants.image_read( ants.get_ants_data('r64'))
    >>> fi = ants.resample_image(fi,(128,128),1,0)
    >>> mi = ants.resample_image(mi,(128,128),1,0)
    >>> mytx = ants.registration(fixed=fi , moving=mi, type_of_transform = ('SyN') )
    >>> warp = ants.image_read( mytx['fwdtransforms'][0] )
    >>> # Use the fast, optimized Python implementation
    >>> dg_py = ants.deformation_gradient( warp, py_based=True )
    >>> dg_rot_py = ants.deformation_gradient( warp, to_rotation=True, py_based=True )
    """
    if not py_based:
        # --- Original C++ based implementation ---
        if ants.is_image(warp_image):
            txuse = NamedTemporaryFile(suffix='.nii.gz', delete=False).name
            ants.image_write(warp_image, txuse)
        else:
            txuse = warp_image
            warp_image=ants.image_read(txuse)
        if not ants.is_image(warp_image):
            raise RuntimeError("antsimage is required")
        writtenimage = NamedTemporaryFile(suffix='.nii.gz', delete=False).name
        dimage = warp_image.split_channels()[0].clone('double')
        dim = dimage.dimension
        tshp = dimage.shape
        args2 = [dim, txuse, writtenimage, int(0), int(0), int(1)]
        processed_args = process_arguments(args2)
        libfn = get_lib_fn('CreateJacobianDeterminantImage')
        libfn(processed_args)
        dg = ants.image_read(writtenimage)
        if to_rotation or to_inverse_rotation:
            newshape = tshp + (dim,dim)
            dg = np.reshape( dg.numpy(), newshape )
            U, s, Vh = np.linalg.svd(dg)
            Z = U @ Vh
            dets = np.linalg.det(Z)
            reflection_mask = dets < 0
            Vh[reflection_mask, -1, :] *= -1
            Z[reflection_mask] = U[reflection_mask] @ Vh[reflection_mask]
            dg = Z
            if to_inverse_rotation:
                dg = np.transpose(dg, axes=(*range(dg.ndim - 2), dg.ndim - 1, dg.ndim - 2))
            newshape = tshp + (dim*dim,)
            dg = np.reshape( dg, newshape )
            dg = ants.from_numpy( dg, has_components=True )
            dg = ants.copy_image_info( dimage, dg )
        import os
        os.remove( writtenimage )
        return dg

    # --- Optimized Python/NumPy Implementation ---
    if py_based:
        if not ants.is_image(warp_image):
            raise RuntimeError("antsimage is required")
        dim = warp_image.dimension
        tshp = warp_image.shape
        tdir = warp_image.direction
        spc = warp_image.spacing
        warpnp = warp_image.numpy()
        gradient_list = [np.gradient(warpnp[..., k], *spc, axis=range(dim)) for k in range(dim)]
        # This correctly calculates J.T, where dg[..., i, j] = d(u_j)/d(x_i)
        dg = np.stack([np.stack(grad_k, axis=-1) for grad_k in gradient_list], axis=-1)
        dg = (tdir @ dg).swapaxes(-1, -2)
        dg += np.eye(dim)
        if to_rotation or to_inverse_rotation:
            U, s, Vh = np.linalg.svd(dg)
            Z = U @ Vh
            dets = np.linalg.det(Z)
            reflection_mask = dets < 0
            Vh[reflection_mask, -1, :] *= -1
            Z[reflection_mask] = U[reflection_mask] @ Vh[reflection_mask]
            dg = Z
            if to_inverse_rotation:
                dg = np.transpose(dg, axes=(*range(dg.ndim - 2), dg.ndim - 1, dg.ndim - 2))
        new_shape = tshp + (dim * dim,)
        dg_reshaped = np.reshape(dg, new_shape)
        return ants.from_numpy(dg_reshaped, origin=warp_image.origin,
                            spacing=warp_image.spacing, direction=warp_image.direction,
                            has_components=True)




def create_jacobian_determinant_image(domain_image, tx, do_log=False, geom=False):
    """
    Compute the jacobian determinant from a transformation file

    ANTsR function: `createJacobianDeterminantImage`

    Arguments
    ---------
    domain_image : ANTsImage
        image that defines transformation domain

    tx : string
        deformation transformation file name

    do_log : boolean
        return the log jacobian

    geom : boolean
        use the geometric jacobian calculation (boolean)

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data('r16'))
    >>> mi = ants.image_read( ants.get_ants_data('r64'))
    >>> fi = ants.resample_image(fi,(128,128),1,0)
    >>> mi = ants.resample_image(mi,(128,128),1,0)
    >>> mytx = ants.registration(fixed=fi , moving=mi, type_of_transform = ('SyN') )
    >>> jac = ants.create_jacobian_determinant_image(fi,mytx['fwdtransforms'][0],1)
    """
    dim = domain_image.dimension
    if ants.is_image(tx):
        txuse = NamedTemporaryFile(suffix='.nii.gz', delete=False).name
        ants.image_write(tx, txuse)
    else:
        txuse = tx
    #args = [dim, txuse, do_log]
    dimage = domain_image.clone('double')
    args2 = [dim, txuse, dimage, int(do_log), int(geom)]
    processed_args = process_arguments(args2)
    libfn = get_lib_fn('CreateJacobianDeterminantImage')
    libfn(processed_args)
    jimage = args2[2].clone('float')

    return jimage




"""
Test utils module
"""

import os
import unittest
from tempfile import TemporaryDirectory

from common import run_tests

import math
import numpy.testing as nptest
import numpy as np
import pandas as pd
try:
    import SimpleITK as sitk
except ImportError:
    sitk = None

import ants


class TestModule_iMath(unittest.TestCase):

    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16'))
        #img3d = ants.image_read(ants.get_ants_data('mni')).resample_image((2,2,2))
        self.imgs = [img2d]

    def tearDown(self):
        pass

    def test_iMath_example(self):
        img = ants.image_read(ants.get_ants_data('r16'))
        img2 = ants.iMath(img, 'Canny', 1, 5, 12)

    def test_iMath_options(self):
        for img in self.imgs:
            # Grayscale dilation
            img_gd = ants.iMath(img, 'GD', 3)
            self.assertTrue(ants.image_physical_space_consistency(img_gd, img))
            self.assertTrue(not ants.allclose(img_gd, img))

            # Grayscale erosion
            img_tx = ants.iMath(img, 'GE', 3)
            self.assertTrue(ants.image_physical_space_consistency(img_tx, img))
            self.assertFalse(ants.allclose(img_tx, img))

            # Morphological dilation
            img_tx = ants.iMath(img, 'MD', 3)
            self.assertTrue(ants.image_physical_space_consistency(img_tx, img))
            self.assertFalse(ants.allclose(img_tx, img))

            # Morphological erosion
            img_tx = ants.iMath(img, 'ME', 3)
            self.assertTrue(ants.image_physical_space_consistency(img_tx, img))
            self.assertFalse(ants.allclose(img_tx, img))

            # Morphological closing
            img_tx = ants.iMath(img, 'MC', 3)
            self.assertTrue(ants.image_physical_space_consistency(img_tx, img))
            self.assertTrue(not ants.allclose(img_tx, img))

            # Morphological opening
            img_tx = ants.iMath(img, 'MO', 3)
            self.assertTrue(ants.image_physical_space_consistency(img_tx, img))
            self.assertTrue(not ants.allclose(img_tx, img))

            # Pad image - ERRORS
            #img_tx = ants.iMath(img, 'PadImage', 2)

            # PeronaMalik - ERRORS
            #img_tx = ants.iMath(img, 'PeronaMalik', 10, 0.5)

            # Maurer Distance
            img_tx = ants.iMath(img, 'MaurerDistance')
            self.assertTrue(ants.image_physical_space_consistency(img_tx, img))
            self.assertTrue(not ants.allclose(img_tx, img))

            # Danielsson Distance Map
            #img_tx = ants.iMath(img, 'DistanceMap')
            #self.assertTrue(ants.image_physical_space_consistency(img_tx, img))
            #self.assertTrue(not ants.allclose(img_tx, img))

            # Image gradient
            img_tx = ants.iMath(img, 'Grad', 1)
            self.assertTrue(ants.image_physical_space_consistency(img_tx, img))
            self.assertTrue(not ants.allclose(img_tx, img))

            # Laplacian
            img_tx = ants.iMath(img, 'Laplacian', 1, 1)
            self.assertTrue(ants.image_physical_space_consistency(img_tx, img))
            self.assertTrue(not ants.allclose(img_tx, img))

class TestModule_bias_correction(unittest.TestCase):
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data("r16"))
        img3d = ants.image_read(ants.get_ants_data("mni")).resample_image((2, 2, 2))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_n3_bias_field_correction_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image_n3 = ants.n3_bias_field_correction(image)

    def test_n3_bias_field_correction(self):
        for img in self.imgs:
            image_n3 = ants.n3_bias_field_correction(img)

    def test_n4_bias_field_correction_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image_n4 = ants.n4_bias_field_correction(image)

        # spline param list
        image_n4 = ants.n4_bias_field_correction(image, spline_param=(10, 10))

        # image not float
        image_ui = image.clone("unsigned int")
        image_n4 = ants.n4_bias_field_correction(image_ui)

        # weight mask
        mask = ants.image_clone(image > image.mean(), pixeltype="float")
        ants.n4_bias_field_correction(image, weight_mask=mask)

        # len(spline_param) != img.dimension
        with self.assertRaises(Exception):
            ants.n4_bias_field_correction(image, spline_param=(10, 10, 10))

        # weight mask not ANTsImage
        with self.assertRaises(Exception):
            ants.n4_bias_field_correction(image, weight_mask=0.4)

    def test_n4_bias_field_correction(self):
        # def n4_bias_field_correction(image, mask=None, shrink_factor=4,
        #                     convergence={'iters':[50,50,50,50], 'tol':1e-07},
        #                     spline_param=200, verbose=False, weight_mask=None):
        for img in self.imgs:
            image_n4 = ants.n4_bias_field_correction(img)
            mask = ants.image_clone(img > img.mean(), pixeltype="float")
            image_n4 = ants.n4_bias_field_correction(img, mask=mask)

    def test_abp_n4_example(self):
        img = ants.image_read(ants.get_ants_data("r16"))
        img = ants.iMath(img, "Normalize") * 255.0
        img2 = ants.abp_n4(img)

    def test_abp_n4(self):
        # def abp_n4(image, intensity_truncation=(0.025,0.975,256), mask=None, usen3=False):
        for img in self.imgs:
            img2 = ants.abp_n4(img)
            img3 = ants.abp_n4(img, usen3=True)

        # intensity trunction doesnt have three values
        with self.assertRaises(Exception):
            img = self.imgs[0]
            ants.abp_n4(img, intensity_truncation=(1, 2))


class TestModule_channels(unittest.TestCase):
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data("r16"))
        img3d = ants.image_read(ants.get_ants_data("mni"))
        self.imgs = [img2d, img3d]
        arr2d = np.random.randn(69, 70, 4).astype("float32")
        arr3d = np.random.randn(69, 70, 71, 4).astype("float32")
        self.comparrs = [arr2d, arr3d]

    def tearDown(self):
        pass

    def test_merge_channels(self):
        for img in self.imgs:
            imglist = [img.clone(), img.clone()]
            merged_img = ants.merge_channels(imglist)

            self.assertEqual(merged_img.shape, img.shape)
            self.assertEqual(merged_img.components, len(imglist))
            nptest.assert_allclose(merged_img.numpy()[..., 0], imglist[0].numpy())

            imglist = [img.clone(), img.clone(), img.clone(), img.clone()]
            merged_img = ants.merge_channels(imglist)

            self.assertEqual(merged_img.shape, img.shape)
            self.assertEqual(merged_img.components, len(imglist))
            nptest.assert_allclose(merged_img.numpy()[..., 0], imglist[-1].numpy())

        for comparr in self.comparrs:
            imglist = [
                ants.from_numpy(comparr[..., i].copy())
                for i in range(comparr.shape[-1])
            ]
            merged_img = ants.merge_channels(imglist)
            for i in range(comparr.shape[-1]):
                nptest.assert_allclose(merged_img.numpy()[..., i], comparr[..., i])


class TestModule_crop_image(unittest.TestCase):
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data("r16"))
        img3d = ants.image_read(ants.get_ants_data("mni")).resample_image((2, 2, 2))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_crop_image_example(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        cropped = ants.crop_image(fi)
        cropped = ants.crop_image(fi, fi, 100)

        # image not float type
        cropped = ants.crop_image(fi.clone("unsigned int"))

        # label image not float
        cropped = ants.crop_image(fi, fi.clone("unsigned int"), 100)

        # channel image
        fi = ants.image_read( ants.get_ants_data('r16') )
        cropped = ants.crop_image(fi)
        fi2 = ants.merge_channels([fi,fi])
        cropped2 = ants.crop_image(fi2)

        self.assertEqual(cropped.shape, cropped2.shape)

    def test_crop_indices_example(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        cropped = ants.crop_indices(fi, (10, 10), (100, 100))
        cropped = ants.smooth_image(cropped, 5)
        decropped = ants.decrop_image(cropped, fi)

        # image not float
        cropped = ants.crop_indices(fi.clone("unsigned int"), (10, 10), (100, 100))

        # image dim not equal to indices
        with self.assertRaises(Exception):
            cropped = ants.crop_indices(fi, (10, 10, 10), (100, 100))
            cropped = ants.crop_indices(fi, (10, 10), (100, 100, 100))

        # vector images
        fi = ants.image_read( ants.get_ants_data("r16"))
        cropped = ants.crop_indices( fi, (10,10), (100,100) )
        fi2 = ants.merge_channels([fi,fi])
        cropped2 = ants.crop_indices( fi, (10,10), (100,100) )
        self.assertEqual(cropped.shape, cropped2.shape)

    def test_decrop_image_example(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        mask = ants.get_mask(fi)
        cropped = ants.crop_image(fi, mask, 1)
        cropped = ants.smooth_image(cropped, 1)
        decropped = ants.decrop_image(cropped, fi)

        # image not float
        cropped = ants.crop_image(fi.clone("unsigned int"), mask, 1)

        # full image not float
        cropped = ants.crop_image(fi, mask.clone("unsigned int"), 1)

class TestModule_pad_image(unittest.TestCase):
    def setUp(self):
        self.img2d = ants.image_read(ants.get_ants_data("r16"))
        self.img3d = ants.image_read(ants.get_ants_data("mni")).resample_image((2, 2, 2))

    def tearDown(self):
        pass

    def test_pad_image_example(self):
        img = self.img2d.clone()
        img.set_origin((0, 0))
        img.set_spacing((2, 2))
        # isotropic pad via pad_width
        padded = ants.pad_image(img, pad_width=(10, 10))
        self.assertEqual(padded.shape, (img.shape[0] + 10, img.shape[1] + 10))
        for img_orig_elem, pad_orig_elem in zip(img.origin, padded.origin):
            self.assertAlmostEqual(pad_orig_elem, img_orig_elem - 10, places=3)

        img = self.img2d.clone()
        img.set_origin((0, 0))
        img.set_spacing((2, 2))
        img = ants.resample_image(img, (128,128), 1, 1)
        # isotropic pad via shape
        padded = ants.pad_image(img, shape=(160,160))
        self.assertSequenceEqual(padded.shape, (160, 160))

        img = self.img3d.clone()
        img = ants.resample_image(img, (128,160,96), 1, 1)
        padded = ants.pad_image(img)
        self.assertSequenceEqual(padded.shape, (160, 160, 160))

        # pad only on superior side
        img = self.img3d.clone()
        padded = ants.pad_image(img, pad_width=[(0,4),(0,8),(0,12)])
        self.assertSequenceEqual(padded.shape, (img.shape[0] + 4, img.shape[1] + 8, img.shape[2] + 12))
        self.assertSequenceEqual(img.origin, padded.origin)


class TestModule_denoise_image(unittest.TestCase):
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data("r16"))
        img3d = ants.image_read(ants.get_ants_data("mni")).resample_image((2, 2, 2))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_denoise_image_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        # add fairly large salt and pepper noise
        imagenoise = image + np.random.randn(*image.shape).astype("float32") * 5
        imagedenoise = ants.denoise_image(imagenoise, image > image.mean())


class TestModule_fit_bspline_object_to_scattered_data(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_fit_bspline_object_to_scattered_data_1d_curve_example(self):
        x = np.linspace(-4, 4, num=100) + np.random.uniform(-0.1, 0.1, 100)
        u = np.linspace(0, 1.0, num=len(x))
        scattered_data = np.expand_dims(u, axis=-1)
        parametric_data = np.expand_dims(u, axis=-1)
        spacing = 1/(len(x)-1) * 1.0
        bspline_curve = ants.fit_bspline_object_to_scattered_data(
            scattered_data, parametric_data,
            parametric_domain_origin=[0.0], parametric_domain_spacing=[spacing],
            parametric_domain_size=[len(x)], is_parametric_dimension_closed=None,
            number_of_fitting_levels=5, mesh_size=1)

    def test_fit_bspline_object_to_scattered_data_2d_curve_example(self):
        x = np.linspace(-4, 4, num=100)
        y = np.exp(-np.multiply(x, x)) + np.random.uniform(-0.1, 0.1, len(x))
        u = np.linspace(0, 1.0, num=len(x))
        scattered_data = np.column_stack((x, y))
        parametric_data = np.expand_dims(u, axis=-1)
        spacing = 1/(len(x)-1) * 1.0
        bspline_curve = ants.fit_bspline_object_to_scattered_data(
            scattered_data, parametric_data,
            parametric_domain_origin=[0.0], parametric_domain_spacing=[spacing],
            parametric_domain_size=[len(x)], is_parametric_dimension_closed=None,
            number_of_fitting_levels=5, mesh_size=1)

    def test_fit_bspline_object_to_scattered_data_3d_curve_example(self):
        x = np.linspace(-4, 4, num=100)
        y = np.exp(-np.multiply(x, x)) + np.random.uniform(-0.1, 0.1, len(x))
        z = np.exp(-np.multiply(x, y)) + np.random.uniform(-0.1, 0.1, len(x))
        u = np.linspace(0, 1.0, num=len(x))
        scattered_data = np.column_stack((x, y))
        parametric_data = np.expand_dims(u, axis=-1)
        spacing = 1/(len(x)-1) * 1.0
        bspline_curve = ants.fit_bspline_object_to_scattered_data(
            scattered_data, parametric_data,
            parametric_domain_origin=[0.0], parametric_domain_spacing=[spacing],
            parametric_domain_size=[len(x)], is_parametric_dimension_closed=None,
            number_of_fitting_levels=5, mesh_size=1)

    def test_fit_bspline_object_to_scattered_data_2d_scalar_field_example(self):
        number_of_random_points = 10000
        img = ants.image_read( ants.get_ants_data("r16"))
        img_array = img.numpy()
        row_indices = np.random.choice(range(2, img_array.shape[0]), number_of_random_points)
        col_indices = np.random.choice(range(2, img_array.shape[1]), number_of_random_points)
        scattered_data = np.zeros((number_of_random_points, 1))
        parametric_data = np.zeros((number_of_random_points, 2))
        for i in range(number_of_random_points):
            scattered_data[i,0] = img_array[row_indices[i], col_indices[i]]
            parametric_data[i,0] = row_indices[i]
            parametric_data[i,1] = col_indices[i]
        bspline_img = ants.fit_bspline_object_to_scattered_data(
            scattered_data, parametric_data,
            parametric_domain_origin=[0.0, 0.0],
            parametric_domain_spacing=[1.0, 1.0],
            parametric_domain_size = img.shape,
            number_of_fitting_levels=7, mesh_size=1)

    def test_fit_bspline_object_to_scattered_data_3d_scalar_field_example(self):
        number_of_random_points = 10000
        img = ants.image_read( ants.get_ants_data("mni"))
        img_array = img.numpy()
        row_indices = np.random.choice(range(2, img_array.shape[0]), number_of_random_points)
        col_indices = np.random.choice(range(2, img_array.shape[1]), number_of_random_points)
        dep_indices = np.random.choice(range(2, img_array.shape[2]), number_of_random_points)
        scattered_data = np.zeros((number_of_random_points, 1))
        parametric_data = np.zeros((number_of_random_points, 3))
        for i in range(number_of_random_points):
            scattered_data[i,0] = img_array[row_indices[i], col_indices[i], dep_indices[i]]
            parametric_data[i,0] = row_indices[i]
            parametric_data[i,1] = col_indices[i]
            parametric_data[i,2] = dep_indices[i]
        bspline_img = ants.fit_bspline_object_to_scattered_data(
            scattered_data, parametric_data,
            parametric_domain_origin=[0.0, 0.0, 0.0],
            parametric_domain_spacing=[1.0, 1.0, 1.0],
            parametric_domain_size = img.shape,
            number_of_fitting_levels=5, mesh_size=1)

    def test_fit_bspline_object_to_scattered_data_2d_displacement_field_example(self):
            img = ants.image_read( ants.get_ants_data("r16"))
            # smooth a single vector in the middle of the image
            scattered_data = np.reshape(np.asarray((10.0, 10.0)), (1, 2))
            parametric_data = np.reshape(np.asarray(
                (scattered_data[0, 0]/img.shape[0], scattered_data[0, 1]/img.shape[1])), (1, 2))
            bspline_field = ants.fit_bspline_object_to_scattered_data(
                scattered_data, parametric_data,
                parametric_domain_origin=[0.0, 0.0],
                parametric_domain_spacing=[1.0, 1.0],
                parametric_domain_size = img.shape,
                number_of_fitting_levels=7, mesh_size=1)

    def test_fit_bspline_object_to_scattered_data_3d_displacement_field_example(self):
            img = ants.image_read( ants.get_ants_data("mni"))
            # smooth a single vector in the middle of the image
            scattered_data = np.reshape(np.asarray((10.0, 10.0, 10.0)), (1, 3))
            parametric_data = np.reshape(np.asarray(
                (scattered_data[0, 0]/img.shape[0],
                 scattered_data[0, 1]/img.shape[1],
                 scattered_data[0, 2]/img.shape[2])), (1, 3))
            bspline_field = ants.fit_bspline_object_to_scattered_data(
                scattered_data, parametric_data,
                parametric_domain_origin=[0.0, 0.0, 0.0],
                parametric_domain_spacing=[1.0, 1.0, 1.0],
                parametric_domain_size = img.shape,
                number_of_fitting_levels=5, mesh_size=1)


class TestModule_get_ants_data(unittest.TestCase):
    def test_get_ants_data(self):
        for dataname in ants.get_ants_data(None):
            if dataname.endswith("nii.gz"):
                img = ants.image_read(dataname)

#    def test_get_ants_data_files(self):
#        datanames = ants.get_ants_data(None)
#        self.assertTrue(isinstance(datanames, list))
#        self.assertTrue(len(datanames) > 0)

    def test_get_ants_data2(self):
        for dataname in ants.get_data(None):
            if dataname.endswith("nii.gz"):
                img = ants.image_read(dataname)

#    def test_get_ants_data_files2(self):
#        datanames = ants.get_data(None)
#        self.assertTrue(isinstance(datanames, list))
#        self.assertTrue(len(datanames) > 0)


class TestModule_get_centroids(unittest.TestCase):
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data("r16"))
        img3d = ants.image_read(ants.get_ants_data("mni")).resample_image((2, 2, 2))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_get_centroids_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image = ants.threshold_image(image, 90, 120)
        image = ants.label_clusters(image, 10)
        cents = ants.get_centroids(image)


class TestModule_get_mask(unittest.TestCase):
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data("r16"))
        img3d = ants.image_read(ants.get_ants_data("mni")).resample_image((2, 2, 2))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_get_mask_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        mask = ants.get_mask(image)


class TestModule_get_neighborhood(unittest.TestCase):
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data("r16"))
        img3d = ants.image_read(ants.get_ants_data("mni")).resample_image((2, 2, 2))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_get_neighborhood_in_mask_example(self):
        img = ants.image_read(ants.get_ants_data("r16"))
        mask = ants.get_mask(img)
        mat = ants.get_neighborhood_in_mask(img, mask, radius=(2, 2))

        # image not ANTsImage
        with self.assertRaises(Exception):
            ants.get_neighborhood_in_mask(2, mask, radius=(2, 2))
        # mask not ANTsImage
        with self.assertRaises(Exception):
            ants.get_neighborhood_in_mask(img, 2, radius=(2, 2))
        # radius not right length
        with self.assertRaises(Exception):
            ants.get_neighborhood_in_mask(img, 2, radius=(2, 2, 2, 2))

        # radius is just a float/int
        mat = ants.get_neighborhood_in_mask(img, mask, radius=2)

        # boundary condition == 'image'
        mat = ants.get_neighborhood_in_mask(
            img, mask, radius=(2, 2), boundary_condition="image"
        )
        # boundary condition == 'mean'
        mat = ants.get_neighborhood_in_mask(
            img, mask, radius=(2, 2), boundary_condition="mean"
        )

        # spatial info
        mat = ants.get_neighborhood_in_mask(img, mask, radius=(2, 2), spatial_info=True)

        # get_gradient
        mat = ants.get_neighborhood_in_mask(img, mask, radius=(2, 2), get_gradient=True)

    def test_get_neighborhood_at_voxel_example(self):
        img = ants.image_read(ants.get_ants_data("r16"))
        center = (2, 2)
        radius = (3, 3)
        retval = ants.get_neighborhood_at_voxel(img, center, radius)

        # image not ANTsImage
        with self.assertRaises(Exception):
            ants.get_neighborhood_at_voxel(2, center, radius)
        # wrong center length
        with self.assertRaises(Exception):
            ants.get_neighborhood_at_voxel(img, (2, 2, 2, 2, 2), radius)
        # wrong radius length
        with self.assertRaises(Exception):
            ants.get_neighborhood_at_voxel(img, center, (2, 2, 2, 2))


class TestModule_image_similarity(unittest.TestCase):
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data("r16"))
        img3d = ants.image_read(ants.get_ants_data("mni")).resample_image((2, 2, 2))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_image_similarity_example(self):
        x = ants.image_read(ants.get_ants_data("r16"))
        y = ants.image_read(ants.get_ants_data("r30"))
        metric = ants.image_similarity(x, y, metric_type="MeanSquares")
        self.assertTrue(metric > 0)


class TestModule_image_to_cluster_images(unittest.TestCase):
    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data("r16"))
        img3d = ants.image_read(ants.get_ants_data("mni")).resample_image((2, 2, 2))
        self.imgs = [img2d, img3d]

    def tearDown(self):
        pass

    def test_image_to_cluster_images_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image = ants.threshold_image(image, 1, 1e15)
        image_cluster_list = ants.image_to_cluster_images(image)


class TestModule_label_clusters(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_label_clusters_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        timageFully = ants.label_clusters(image, 10, 128, 150, True)
        timageFace = ants.label_clusters(image, 10, 128, 150, False)


class TestModule_label_image_centroids(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_label_image_centroids(self):
        image = ants.from_numpy(
            np.asarray([[[0, 2], [1, 3]], [[4, 6], [5, 7]]]).astype("float32")
        )
        labels = ants.label_image_centroids(image)
        self.assertEqual(len(labels['labels']), 7)

        # Test non-sequential labels
        image = ants.from_numpy(
            np.asarray([[[0, 2], [2, 2]], [[2, 0], [5, 0]]]).astype("float32")
        )

        labels = ants.label_image_centroids(image)
        self.assertTrue(len(labels['labels']) == 2)
        self.assertTrue(labels['labels'][1] == 5)
        self.assertTrue(np.allclose(labels['vertices'][0], [0.5 , 0.5 , 0.25], atol=1e-5))
        # With convex = False, the centroid position should change
        labels = ants.label_image_centroids(image, convex=False)
        self.assertTrue(np.allclose(labels['vertices'][0], [1.0, 1.0, 0.0], atol=1e-5))
        # single point unchanged
        self.assertTrue(np.allclose(labels['vertices'][1], [0.0, 1.0, 1.0], atol=1e-5))


class TestModule_label_overlap_measures(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_label_overlap_measures(self):
        r16 = ants.image_read( ants.get_ants_data('r16') )
        r64 = ants.image_read( ants.get_ants_data('r64') )
        s16 = ants.kmeans_segmentation( r16, 3 )['segmentation']
        s64 = ants.kmeans_segmentation( r64, 3 )['segmentation']
        stats = ants.label_overlap_measures(s16, s64)


class TestModule_label_stats(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_label_stats_example(self):
        image = ants.image_read(ants.get_ants_data("r16"), 2)
        image = ants.resample_image(image, (64, 64), 1, 0)
        segs1 = ants.kmeans_segmentation(image, 3)
        stats = ants.label_stats(image, segs1["segmentation"])


class TestModule_labels_to_matrix(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_labels_to_matrix_example(self):
        fi = ants.image_read(ants.get_ants_data("r16")).resample_image((60, 60), 1, 0)
        mask = ants.get_mask(fi)
        labs = ants.kmeans_segmentation(fi, 3)["segmentation"]
        labmat = ants.labels_to_matrix(labs, mask)


class TestModule_make_points_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_make_points_image_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        points = np.array([[102,  76],[134, 129]])
        points_image = ants.make_points_image(points, image, radius=5)
        stats = ants.label_stats(image, points_image)
        self.assertTrue(np.allclose(stats['Volume'].to_numpy()[1:3], 97.0, atol=1e-5))
        self.assertTrue(np.allclose(stats['x'].to_numpy()[1:3], points[:,0], atol=1e-5))
        self.assertTrue(np.allclose(stats['y'].to_numpy()[1:3], points[:,1], atol=1e-5))

        points = np.array([[102,  76,  50],[134, 129,  50]])
        # Shouldn't allow 3D points on a 2D image
        with self.assertRaises(Exception):
            points_image = ants.make_points_image(image, points, radius=3)


class TestModule_mask_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_mask_image_example(self):
        myimage = ants.image_read(ants.get_ants_data("r16"))
        mask = ants.get_mask(myimage)
        myimage_mask = ants.mask_image(myimage, mask, (2, 3))
        seg = ants.kmeans_segmentation(myimage, 3)
        myimage_mask = ants.mask_image(myimage, seg["segmentation"], (1, 3))


class TestModule_mni2tal(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_mni2tal_example(self):
        ants.mni2tal((10, 12, 14))


class TestModule_morphology(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_morphology(self):
        fi = ants.image_read(ants.get_ants_data("r16"))
        mask = ants.get_mask(fi)
        dilated_ball = ants.morphology(
            mask, operation="dilate", radius=3, mtype="binary", shape="ball"
        )
        eroded_box = ants.morphology(
            mask, operation="erode", radius=3, mtype="binary", shape="box"
        )
        opened_annulus = ants.morphology(
            mask,
            operation="open",
            radius=5,
            mtype="binary",
            shape="annulus",
            thickness=2,
        )

        ## --- ##
        mtype = "binary"
        mask = mask.clone()
        for operation in {"dilate", "erode", "open", "close"}:
            for shape in {"ball", "box", "cross", "annulus", "polygon"}:
                morph = ants.morphology(
                    mask, operation=operation, radius=3, mtype=mtype, shape=shape
                )
                self.assertTrue(isinstance(morph, ants.core.ants_image.ANTsImage))
        # invalid operation
        with self.assertRaises(Exception):
            ants.morphology(
                mask, operation="invalid-operation", radius=3, mtype=mtype, shape=shape
            )

        mtype = "grayscale"
        img = ants.image_read(ants.get_ants_data("r16"))
        for operation in {"dilate", "erode", "open", "close"}:
            morph = ants.morphology(img, operation=operation, radius=3, mtype=mtype)
            self.assertTrue(isinstance(morph, ants.core.ants_image.ANTsImage))
        # invalid operation
        with self.assertRaises(Exception):
            ants.morphology(
                img, operation="invalid-operation", radius=3, mtype="grayscale"
            )

        # invalid morphology type
        with self.assertRaises(Exception):
            ants.morphology(
                img, operation="dilate", radius=3, mtype="invalid-morphology"
            )

        # invalid shape
        with self.assertRaises(Exception):
            ants.morphology(
                mask,
                operation="dilate",
                radius=3,
                mtype="binary",
                shape="invalid-shape",
            )


class TestModule_ndimage_to_list(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_ndimage_to_list(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        image2 = ants.image_read(ants.get_ants_data("r64"))
        ants.set_spacing(image, (2, 2))
        ants.set_spacing(image2, (2, 2))
        imageTar = ants.make_image((*image2.shape, 2))
        ants.set_spacing(imageTar, (2, 2, 2))
        image3 = ants.list_to_ndimage(imageTar, [image, image2])
        self.assertEqual(image3.dimension, 3)
        ants.set_direction(image3, np.eye(3) * 2)
        images_unmerged = ants.ndimage_to_list(image3)
        self.assertEqual(len(images_unmerged), 2)
        self.assertEqual(images_unmerged[0].dimension, 2)


class TestModule_simulate_displacement_field(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_simulate_exponential_displacement_field_2d(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        exp_field = ants.simulate_displacement_field(image, field_type="exponential")

    def test_simulate_bspline_displacement_field_2d(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        bsp_field = ants.simulate_displacement_field(image, field_type="bspline")


class TestModule_smooth_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_smooth_image_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        simage = ants.smooth_image(image, (1.2, 1.5))


class TestModule_threshold_image(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_threshold_image_example(self):
        image = ants.image_read(ants.get_ants_data("r16"))
        timage = ants.threshold_image(image, 0.5, 1e15)


class TestModule_weingarten_image_curvature(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_weingarten_image_curvature_example(self):
        image = ants.image_read(ants.get_ants_data("mni")).resample_image((3, 3, 3))
        imagecurv = ants.weingarten_image_curvature(image)

        image2 = ants.image_read(ants.get_ants_data("r16")).resample_image((2, 2))
        imagecurv2 = ants.weingarten_image_curvature(image2)


class TestModule_scalar_rgb_vector(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test1(self):
        import ants
        import numpy as np

        # this fails because ConvertScalarImageToRGB is not wrapped
        img = ants.image_read(ants.get_data("r16"), pixeltype="unsigned char")
        # img_rgb = img.scalar_to_rgb()
        # img_vec = img_rgb.rgb_to_vector()

        # rgb_arr = img_rgb.numpy()
        # vec_arr = img_vec.numpy()
        print(np.allclose(img.numpy(), img.numpy()))

    def test2(self):
        import ants
        import numpy as np

        rgb_img = ants.from_numpy(
            np.random.randint(0, 255, (20, 20, 3)).astype("uint8"), is_rgb=True
        )
        vec_img = rgb_img.rgb_to_vector()
        print(ants.allclose(rgb_img, vec_img))

        vec_img = ants.from_numpy(
            np.random.randint(0, 255, (20, 20, 3)).astype("uint8"), has_components=True
        )
        rgb_img = vec_img.vector_to_rgb()
        print(ants.allclose(rgb_img, vec_img))


class TestRandom(unittest.TestCase):
    def setUp(self):
        pass
    def tearDown(self):
        pass

    def test_bspline_field(self):
        points = np.array([[-50, -50]])
        deltas = np.array([[10, 10]])
        bspline_field = ants.fit_bspline_displacement_field(
        displacement_origins=points, displacements=deltas,
        origin=[0.0, 0.0], spacing=[1.0, 1.0], size=[100, 100],
        direction=np.array([[-1, 0], [0, -1]]),
        number_of_fitting_levels=4, mesh_size=(1, 1))

    def test_quantile(self):
        img = ants.image_read(ants.get_data('r16'))
        ants.rank_intensity(img)

    def test_ilr(self):
        nsub = 20
        mu, sigma = 0, 1
        outcome = np.random.normal( mu, sigma, nsub )
        covar = np.random.normal( mu, sigma, nsub )
        mat = np.random.normal( mu, sigma, (nsub, 500 ) )
        mat2 = np.random.normal( mu, sigma, (nsub, 500 ) )
        data = {'covar':covar,'outcome':outcome}
        df = pd.DataFrame( data )
        vlist = { "mat1": mat, "mat2": mat2 }
        myform = " outcome ~ covar * mat1 "
        result = ants.ilr( df, vlist, myform)
        myform = " mat2 ~ covar + mat1 "
        result = ants.ilr( df, vlist, myform)

    def test_quantile(self):
        img = ants.image_read(ants.get_data('r16'))
        ants.quantile(img, 0.5)
        ants.quantile(img, (0.5, 0.75))

    def test_bandpass(self):
        brainSignal = np.random.randn( 400, 1000 )
        tr = 1
        filtered = ants.bandpass_filter_matrix( brainSignal, tr = tr )

    def test_compcorr(self):
        cc = ants.compcor( ants.image_read(ants.get_ants_data("ch2")) )

    def test_histogram_match(self):
        src_img = ants.image_read(ants.get_data('r16'))
        ref_img = ants.image_read(ants.get_data('r64'))
        src_ref = ants.histogram_match_image(src_img, ref_img)

        src_img = ants.image_read(ants.get_data('r16'))
        ref_img = ants.image_read(ants.get_data('r64'))
        src_ref = ants.histogram_match_image2(src_img, ref_img)

    def test_averaging(self):
        x0=[ ants.get_data('r16'), ants.get_data('r27'), ants.get_data('r62'), ants.get_data('r64') ]
        x1=[]
        for k in range(len(x0)):
            x1.append( ants.image_read( x0[k] ) )
        avg=ants.average_images(x0)
        avg1=ants.average_images(x1)
        avg2=ants.average_images(x1,mask=0)
        avg3=ants.average_images(x1,mask=1,normalize=True)

    def test_n3_2(self):
        image = ants.image_read( ants.get_ants_data('r16') )
        image_n3 = ants.n3_bias_field_correction2(image)

    def test_add_noise(self):
        image = ants.image_read(ants.get_ants_data('r16'))
        noise_image = ants.add_noise_to_image(image, 'additivegaussian', (0.0, 1.0))
        noise_image = ants.add_noise_to_image(image, 'saltandpepper', (0.1, 0.0, 100.0))
        noise_image = ants.add_noise_to_image(image, 'shot', 1.0)
        noise_image = ants.add_noise_to_image(image, 'speckle', 1.0)

    def test_hessian_objectness(self):
        image = ants.image_read(ants.get_ants_data('r16'))
        hessian = ants.hessian_objectness(image)

    def test_thin_plate_spline(self):
        points = np.array([[-50, -50]])
        deltas = np.array([[10, 10]])
        tps_field = ants.fit_thin_plate_spline_displacement_field(
        displacement_origins=points, displacements=deltas,
        origin=[0.0, 0.0], spacing=[1.0, 1.0], size=[100, 100],
        direction=np.array([[-1, 0], [0, -1]]))

    def test_multi_label_morph(self):
        img = ants.image_read(ants.get_data('r16'))
        labels = ants.get_mask(img,1,150) + ants.get_mask(img,151,225) * 2
        labels_dilated = ants.multi_label_morphology(labels, 'MD', 2)
        # should see original label regions preserved in dilated version
        # label N should have mean N and 0 variance
        print(ants.label_stats(labels_dilated, labels))

    def test_hausdorff_distance(self):
        r16 = ants.image_read( ants.get_ants_data('r16') )
        r64 = ants.image_read( ants.get_ants_data('r64') )
        s16 = ants.kmeans_segmentation( r16, 3 )['segmentation']
        s64 = ants.kmeans_segmentation( r64, 3 )['segmentation']
        stats = ants.hausdorff_distance(s16, s64)

    def test_channels_first(self):
        image = ants.image_read(ants.get_ants_data('r16'))
        image2 = ants.image_read(ants.get_ants_data('r16'))
        img3 = ants.merge_channels([image,image2])
        img4 = ants.merge_channels([image,image2], channels_first=True)
        self.assertTrue(np.allclose(img3.numpy()[:,:,0], img4.numpy()[0,:,:]))
        self.assertTrue(np.allclose(img3.numpy()[:,:,1], img4.numpy()[1,:,:]))
        
    def test_polar_decomposition(self):
        # Helper functions for creating known matrices
        def make_known_rotation(theta_deg):
            theta = np.deg2rad(theta_deg)
            R = np.eye(3)
            R[:2, :2] = [[np.cos(theta), -np.sin(theta)],
                        [np.sin(theta),  np.cos(theta)]]
            return R
        def make_known_scaling_matrix(sx, sy, sz):
            return np.diag([sx, sy, sz])        

        # 1. Setup: Create a matrix from a known P and Z.        
        # The key is to multiply them in the order P @ Z.
        P_known = make_known_scaling_matrix(2.5, 1.0, 1.5)
        Z_known = make_known_rotation(42) # Use Z for "orthogonal" part        
        # Construct X using the left decomposition structure: X = P @ Z
        X = P_known @ Z_known

        # 2. Decompose the matrix using the function       
        result = ants.polar_decomposition(X)
        P_est = result["P"]
        Z_est = result["Z"]
        X_reconstructed = result["Xtilde"]

        # 3. Assertions
        # a. Check if the reconstruction is close to the original matrix.
        nptest.assert_almost_equal(X, X_reconstructed)
        # b. Check if the estimated symmetric part is close to the known one.
        nptest.assert_almost_equal(P_known, P_est)
        # c. Check if the estimated orthogonal part is close to the known one.
        nptest.assert_almost_equal(Z_known, Z_est)
        
    def test_convergence_monitoring(self):
        f = [1 / i for i in range(1, 100)]
        convergence = ants.convergence_monitoring(f, window_size=10)
        nptest.assert_almost_equal(convergence, 0.0, decimal=3)

@unittest.skipIf(sitk is None, "SimpleITK is not installed")
class TestModule_sitk_to_ants(unittest.TestCase):
    def setUp(self):
        shape = (32, 24, 16)
        self.img = sitk.GetImageFromArray(np.arange(np.prod(shape), dtype=float).reshape(shape))
        self.img.SetSpacing([0.5, 0.5, 2.0])
        self.img.SetOrigin([1.2, 5.7, -3.4])
        self.img.SetDirection(sitk.VersorTransform([1, 0, 0], 0.5).GetMatrix())

    def tearDown(self):
        pass

    # Basic tests with 3D image
    def test_from_sitk(self):
        ants_img = ants.from_sitk(self.img)

        with TemporaryDirectory() as temp_dir:
            temp_fpath = os.path.join(temp_dir, "img.nrrd")
            ants.image_write(ants_img, temp_fpath)
            img = sitk.ReadImage(temp_fpath)

        nptest.assert_equal(
            sitk.GetArrayViewFromImage(self.img), sitk.GetArrayViewFromImage(img)
        )
        nptest.assert_almost_equal(self.img.GetOrigin(), img.GetOrigin())
        nptest.assert_almost_equal(self.img.GetSpacing(), img.GetSpacing())
        nptest.assert_almost_equal(self.img.GetDirection(), img.GetDirection())


    def test_to_sitk(self):
        with TemporaryDirectory() as temp_dir:
            temp_fpath = os.path.join(temp_dir, "img.nrrd")
            sitk.WriteImage(self.img, temp_fpath)
            ants_img = ants.image_read(temp_fpath)

        img = ants.to_sitk(ants_img)

        nptest.assert_equal(
            sitk.GetArrayViewFromImage(self.img), sitk.GetArrayViewFromImage(img)
        )
        nptest.assert_almost_equal(self.img.GetOrigin(), img.GetOrigin())
        nptest.assert_almost_equal(self.img.GetSpacing(), img.GetSpacing())
        nptest.assert_almost_equal(self.img.GetDirection(), img.GetDirection())


    def test_roundtrip_scalar_vector(self):
        """
        Tests roundtrip conversions between ANTsPy and SimpleITK
        for scalar and vector images in 2D, 3D, and 4D.
        """
        # Load 2D scalar image
        base_2d = ants.image_read(ants.get_ants_data('r16'))
        base_2d = ants.resample_image(base_2d, (48, 64), use_voxels=True, interp_type=1)

        # Set interesting spacing, origin, direction for 2D
        base_2d.set_spacing((1.5, 2.0))
        base_2d.set_origin((1.2, 3.4))
        theta = math.radians(16)
        base_2d.set_direction([
           [math.cos(theta), -math.sin(theta)],
           [math.sin(theta),  math.cos(theta)]
        ])

        base_arr_2d = base_2d.numpy()

        test_cases = []

        # 2D scalar
        img_2d_scalar = ants.from_numpy(
            base_arr_2d,
            has_components=False,
            spacing=base_2d.spacing,
            origin=base_2d.origin,
            direction=base_2d.direction,
        )
        test_cases.append(("2D scalar", img_2d_scalar))

        # 2D vector
        arr_vec = np.stack([base_arr_2d, base_arr_2d], axis=-1)
        img_2d_vector = ants.from_numpy(
            arr_vec,
            has_components=True,
            spacing=base_2d.spacing,
            origin=base_2d.origin,
            direction=base_2d.direction,
        )
        test_cases.append(("2D vector", img_2d_vector))

        # --- 3D setup ---

        # Make 3D volume by stacking slices
        arr_3d = np.stack([base_arr_2d]*5, axis=-1)

        # 3D spacing and origin
        spacing_3d = base_2d.spacing + (2.5,)  # (1.5, 2.0, 2.5)
        origin_3d = base_2d.origin + (5.6,)    # (1.2, 3.4, 5.6)

        # 3D direction: rotation about Z axis
        theta = math.radians(30)  # 30 degrees
        direction_3d = np.array([
            [math.cos(theta), -math.sin(theta), 0],
            [math.sin(theta),  math.cos(theta), 0],
            [0,                 0,                1]
        ])

        # 3D scalar
        img_3d_scalar = ants.from_numpy(
            arr_3d,
            has_components=False,
            spacing=spacing_3d,
            origin=origin_3d,
            direction=direction_3d,
        )
        test_cases.append(("3D scalar", img_3d_scalar))

        # 3D vector
        arr_vec_3d = np.stack([arr_3d, arr_3d], axis=-1)
        img_3d_vector = ants.from_numpy(
            arr_vec_3d,
            has_components=True,
            spacing=spacing_3d,
            origin=origin_3d,
            direction=direction_3d,
        )
        test_cases.append(("3D vector", img_3d_vector))

        # --- 4D setup ---

        # Make 4D volume by stacking 3D volumes
        arr_4d = np.stack([arr_3d]*4, axis=-1)

        # 4D spacing and origin
        spacing_4d = spacing_3d + (0.8,)
        origin_4d = origin_3d + (0.0,)

        # 4D direction matrix is same as 3D in space
        direction_4d = np.eye(4)
        direction_4d[:3, :3] = direction_3d


        # 4D scalar
        img_4d_scalar = ants.from_numpy(
            arr_4d,
            has_components=False,
            spacing=spacing_4d,
            origin=origin_4d,
            direction=direction_4d,
        )
        test_cases.append(("4D scalar", img_4d_scalar))

        # 4D vector
        arr_vec_4d = np.stack([arr_4d, arr_4d], axis=-1)
        img_4d_vector = ants.from_numpy(
            arr_vec_4d,
            has_components=True,
            spacing=spacing_4d,
            origin=origin_4d,
            direction=direction_4d,
        )
        test_cases.append(("4D vector", img_4d_vector))

        # Now test round-trip for each
        for name, ants_img in test_cases:
            print(f"Testing {name}")

            # Convert to SimpleITK and back
            sitk_img = ants.to_sitk(ants_img)
            ants_img2 = ants.from_sitk(sitk_img)

            # Basic checks
            assert ants_img.dimension == ants_img2.dimension, f"Dimension mismatch for {name}"
            assert ants_img.shape == ants_img2.shape, f"Shape mismatch for {name}"
            assert ants_img.components == ants_img2.components, f"Components mismatch for {name}"

            # Pixel value checks (floating point tolerant)
            arr1 = ants_img.numpy()
            arr2 = ants_img2.numpy()
            np.testing.assert_allclose(
                arr1, arr2, rtol=1e-5, atol=1e-8,
                err_msg=f"Pixel data mismatch for {name}"
            )

if __name__ == "__main__":
    run_tests()



"""
Atropos segmentation
"""



__all__ = ['atropos']

import os
import glob
import warnings
from tempfile import mktemp


import ants
from ants.internal import get_lib_fn, get_pointer_string, process_arguments


def atropos(a, x, i='Kmeans[3]', m='[0.2,1x1]', c='[5,0]',
            priorweight=0.25, **kwargs):
    """
    A finite mixture modeling (FMM) segmentation approach with possibilities
    for specifying prior constraints. These prior constraints include the
    specification of a prior label image, prior probability images (one for
    each class), and/or an MRF prior to enforce spatial smoothing of the
    labels. Similar algorithms include FAST and SPM. atropos can also perform
    multivariate segmentation if you pass a list of images in: e.g. a=(img1,img2).

    ANTsR function: `atropos`

    Arguments
    ---------
    a : ANTsImage or list/tuple of ANTsImage types
        One or more scalar images to segment. If priors are not used,
        the intensities of the first image are used to order the classes
        in the segmentation output, from lowest to highest intensity. Otherwise
        the order of the classes is dictated by the order of the prior images.

    x : ANTsImage
        mask image.

    i : string
        initialization usually KMeans[N] for N classes or a list of N prior
        probability images. See Atropos in ANTs for full set of options.

    m : string
        mrf parameters as a string, usually "[smoothingFactor,radius]" where
        smoothingFactor determines the amount of smoothing and radius determines
        the MRF neighborhood, as an ANTs style neighborhood vector eg "1x1x1"
        for a 3D image. The radius must match the dimensionality of the image,
        eg 1x1 for 2D and The default in ANTs is smoothingFactor=0.3 and
        radius=1. See Atropos for more options.

    c : string
        convergence parameters, "[numberOfIterations,convergenceThreshold]".
        A threshold of 0 runs the full numberOfIterations, otherwise Atropos
        tests convergence by comparing the mean maximum posterior probability
        over the whole region of interest defined by the mask x.

    priorweight : scalar
        usually 0 (priors used for initialization only), 0.25 or 0.5.

    kwargs : keyword arguments
        more parameters, see Atropos help in ANTs

    Returns
    -------
    dictionary with the following key/value pairs:
        `segmentation`: ANTsImage
            actually segmented image
        `probabilityimages` : list of ANTsImage types
            one image for each segmentation class

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_ants_data('r16'))
    >>> img = ants.resample_image(img, (64,64), 1, 0)
    >>> mask = ants.get_mask(img)
    >>> ants.atropos( a = img, m = '[0.2,1x1]', c = '[2,0]',  i = 'kmeans[3]', x = mask )
    >>> seg2 = ants.atropos( a = img, m = '[0.2,1x1]', c = '[2,0]', i = seg['probabilityimages'], x = mask, priorweight=0.25 )
    """
    probs = mktemp(prefix='antsr', suffix='prob%02d.nii.gz')
    tdir = probs.replace(os.path.basename(probs),'')
    probsbase = os.path.basename(probs)
    searchpattern = probsbase.replace('%02d', '*')

    ct = 0
    if isinstance(i, (list,tuple)) and (len(i) > 1):
        while ct < len(i):
            probchar = str(ct+1)
            if ct < 9:
                probchar = '0%s' % probchar
            tempfn = probs.replace('%02d', probchar)
            ants.image_write(i[ct], tempfn)
            ct += 1
        i = 'PriorProbabilityImages[%s,%s,%s]' % (str(len(i)), probs, str(priorweight))

    if isinstance(a, (list,tuple)):
        outimg = a[0].clone('unsigned int')
    else:
        outimg = a.clone('unsigned int')

    mydim = outimg.dimension
    outs = '[%s,%s]' % (get_pointer_string(outimg), probs)
    mymask = x.clone('unsigned int')

    if (not isinstance(a, (list,tuple))) or (len(a) == 1):
        myargs = {
            'd': mydim,
            'a': a,
            'm-MULTINAME-0': m,
            'o': outs,
            'c': c,
            'm-MULTINAME-1': m,
            'i': i,
            'x': mymask
        }
        for k, v in kwargs.items():
            myargs[k] = v

    elif isinstance(a, (list, tuple)):
        if len(a) > 6:
            print('more than 6 input images not really supported, using first 6')
            a = a[:6]
        myargs = {
            'd': mydim,
            'm-MULTINAME-0': m,
            'o': outs,
            'c': c,
            'm-MULTINAME-1': m,
            'i': i,
            'x': mymask
        }
        for aa_idx, aa in enumerate(a):
            myargs['a-MULTINAME-%i'%aa_idx] = aa

    processed_args = process_arguments(myargs)
    libfn = get_lib_fn('Atropos')
    retval = libfn(processed_args)

    if retval != 0:
        raise Exception(f"Atropos exited with non-zero status {retval}. Run with verbose=1 to see error messages")

    probsout = glob.glob(os.path.join(tdir,'*'+searchpattern))

    if probsout is None or len(probsout) == 0:
        raise Exception('No Atropos output probability images found. Run with verbose=1 to see error messages')

    probsout.sort()
    probimgs = [ants.image_read(probsout[0])]
    for idx in range(1, len(probsout)):
        probimgs.append(ants.image_read(probsout[idx]))

    outimg = outimg.clone('float')
    return {'segmentation': outimg,
            'probabilityimages': probimgs}



__all__ = ["fit_transform_to_paired_points", 
           "fit_time_varying_transform_to_point_sets"]

import numpy as np
import math
import time

import ants

def fit_transform_to_paired_points(moving_points,
                                   fixed_points,
                                   transform_type="affine",
                                   regularization=1e-6,
                                   domain_image=None,
                                   number_of_fitting_levels=4,
                                   mesh_size=1,
                                   spline_order=3,
                                   enforce_stationary_boundary=True,
                                   displacement_weights=None,
                                   number_of_compositions=10,
                                   composition_step_size=0.5,
                                   sigma=0.0,
                                   convergence_threshold=1e-6,
                                   number_of_time_steps=2,
                                   number_of_integration_steps=100,
                                   rasterize_points=False,
                                   verbose=False
                                  ):
    """
    Estimate a transform from corresponding fixed and moving landmarks.

    ANTsR function: fitTransformToPairedPoints

    Arguments
    ---------
    moving_points : array
        Moving points specified in physical space as a n x d matrix where n is the number
        of points and d is the dimensionality.

    fixed_points : array
        Fixed points specified in physical space as a n x d matrix where n is the number
        of points and d is the dimensionality.

    transform_type : character
        'rigid', 'similarity', "affine', 'bspline', 'tps', 'diffeo', 'syn', or 'time-varying (tv)'.

    regularization : scalar
        Ridge penalty in [0,1] for linear transforms.

    domain_image : ANTs image
        Defines physical domain of the nonlinear transform.  Must be defined for nonlinear
        transforms.

    number_of_fitting_levels : integer
        Integer specifying the number of fitting levels for the B-spline interpolation of the
        displacement field.

    mesh_size : integer or array
        Defines the mesh size at the initial fitting level for the B-spline interpolation of the
        displacement field.

    spline_order : integer
        Spline order of the B-spline displacement field.

    enforce_stationary_boundary : boolean
        Ensure no displacements on the image boundary (B-spline only).

    displacement_weights : array
        Defines the individual weighting of the corresponding scattered data value.  Default = NULL
        meaning all displacements are weighted the same.

    number_of_compositions : integer
        Total number of compositions for the diffeomorphic transforms.

    composition_step_size : scalar
        Scalar multiplication factor of the weighting of the update field for the diffeomorphic transforms.

    sigma : scalar
        Gaussian smoothing standard deviation of the update field (in mm).

    convergence_threshold : scalar
        Composition-based convergence parameter for the diff. transforms using a
        window size of 10 values.

    number_of_time_steps : integer
        Time-varying velocity field parameter.

    number_of_integration_steps : scalar
        Number of steps used for integrating the velocity field.

    rasterize_points : boolean
       Use nearest neighbor rasterization of points for estimating the update
       field (potential speed-up).  Default = False.

    verbose : bool
        Print progress to the screen.

    Returns
    -------

    ANTs transform

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> fixed = np.array([[50.0,50.0],[200.0,50.0],[200.0,200.0]])
    >>> moving = np.array([[50.0,50.0],[50.0,200.0],[200.0,200.0]])
    >>> xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="affine")
    >>> xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="rigid")
    >>> xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="similarity")
    >>> domain_image = ants.image_read(ants.get_ants_data("r16"))
    >>> xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="bspline", domain_image=domain_image, number_of_fitting_levels=5)
    >>> xfrm = ants.fit_transform_to_paired_points(moving, fixed, transform_type="diffeo", domain_image=domain_image, number_of_fitting_levels=6)
    """

    def create_zero_displacement_field(domain_image):
         field_array = np.zeros((*domain_image.shape, domain_image.dimension))
         field = ants.from_numpy(field_array, origin=domain_image.origin,
                 spacing=domain_image.spacing, direction=domain_image.direction,
                 has_components=True)
         return(field)

    def create_zero_velocity_field(domain_image, number_of_time_points=2):
         field_array = np.zeros((*domain_image.shape, number_of_time_points, domain_image.dimension))
         origin = (*domain_image.origin, 0.0)
         spacing = (*domain_image.spacing, 1.0)
         direction = np.eye(domain_image.dimension + 1)
         direction[0:domain_image.dimension,0:domain_image.dimension] = domain_image.direction
         field = ants.from_numpy(field_array, origin=origin, spacing=spacing, direction=direction,
                 has_components=True)
         return(field)

    allowed_transforms = ['rigid', 'affine', 'similarity', 'bspline', 'tps', 'diffeo', 'syn', 'tv', 'time-varying']
    if not transform_type.lower() in allowed_transforms:
        raise ValueError(transform_type + " transform not supported.")

    transform_type = transform_type.lower()

    if domain_image is None and transform_type in ['bspline', 'tps', 'diffeo', 'syn', 'tv', 'time-varying']:
        raise ValueError("Domain image needs to be specified.")

    if not fixed_points.shape == moving_points.shape:
        raise ValueError("Mismatch in the size of the point sets.")

    if regularization > 1:
        regularization = 1
    elif regularization < 0:
        regularization = 0

    number_of_points = fixed_points.shape[0]
    dimensionality = fixed_points.shape[1]

    if transform_type in ['rigid', 'affine', 'similarity']:
        center_fixed = fixed_points.mean(axis=0)
        center_moving = moving_points.mean(axis=0)

        x = fixed_points - center_fixed
        y = moving_points - center_moving

        y_prior = np.concatenate((y, np.ones((number_of_points, 1))), axis=1)

        x11 = np.concatenate((x, np.ones((number_of_points, 1))), axis=1)
        M = x11 * (1.0 - regularization) + regularization * y_prior
        Minv = np.linalg.lstsq(M, y, rcond=None)[0]

        p = ants.polar_decomposition(Minv[0:dimensionality, 0:dimensionality].T)
        A = p['Xtilde']
        translation = Minv[dimensionality,:] + center_moving - center_fixed

        if transform_type in ['rigid', 'similarity']:
            # Kabsch algorithm
            #    http://web.stanford.edu/class/cs273/refs/umeyama.pdf

            C = np.dot(y.T, x)
            x_svd = np.linalg.svd(C * (1.0 - regularization) + np.eye(dimensionality) * regularization)
            x_det = np.linalg.det(np.dot(x_svd[0], x_svd[2]))

            if x_det < 0:
                x_svd[2][dimensionality-1, :] *= -1

            A = np.dot(x_svd[0], x_svd[2])

            if transform_type == 'similarity':
                scaling = (math.sqrt((np.power(y, 2).sum(axis=1) / number_of_points).mean()) /
                           math.sqrt((np.power(x, 2).sum(axis=1) / number_of_points).mean()))
                A = np.dot(A, np.eye(dimensionality) * scaling)

        xfrm = ants.create_ants_transform(matrix=A, translation=translation,
              dimension=dimensionality, center=center_fixed)

        return xfrm

    elif transform_type == "bspline":

        bspline_displacement_field = ants.fit_bspline_displacement_field(
            displacement_origins=fixed_points,
            displacements=moving_points - fixed_points,
            displacement_weights=displacement_weights,
            origin=domain_image.origin,
            spacing=domain_image.spacing,
            size=domain_image.shape,
            direction=domain_image.direction,
            number_of_fitting_levels=number_of_fitting_levels,
            mesh_size=mesh_size,
            spline_order=spline_order,
            enforce_stationary_boundary=enforce_stationary_boundary,
            rasterize_points=rasterize_points)

        xfrm = ants.transform_from_displacement_field(bspline_displacement_field)

        return xfrm

    elif transform_type == "tps":

        tps_displacement_field = ants.fit_thin_plate_spline_displacement_field(
            displacement_origins=fixed_points,
            displacements=moving_points - fixed_points,
            origin=domain_image.origin,
            spacing=domain_image.spacing,
            size=domain_image.shape,
            direction=domain_image.direction)

        xfrm = ants.transform_from_displacement_field(tps_displacement_field)

        return xfrm

    elif transform_type == "diffeo":

        if verbose:
            start_total_time = time.time()

        updated_fixed_points = np.empty_like(fixed_points)
        updated_fixed_points[:] = fixed_points

        total_field = create_zero_displacement_field(domain_image)
        total_field_xfrm = None

        error_values = []
        for i in range(number_of_compositions):

            if verbose:
                start_time = time.time()

            update_field = ants.fit_bspline_displacement_field(
              displacement_origins=updated_fixed_points,
              displacements=moving_points - updated_fixed_points,
              displacement_weights=displacement_weights,
              origin=domain_image.origin,
              spacing=domain_image.spacing,
              size=domain_image.shape,
              direction=domain_image.direction,
              number_of_fitting_levels=number_of_fitting_levels,
              mesh_size=mesh_size,
              spline_order=spline_order,
              enforce_stationary_boundary=True,
              rasterize_points=rasterize_points
            )

            update_field = update_field * composition_step_size
            if sigma > 0:
                update_field = ants.smooth_image(update_field, sigma)

            total_field = ants.compose_displacement_fields(update_field, total_field)
            total_field_xfrm = ants.transform_from_displacement_field(total_field)

            if i < number_of_compositions - 1:
                for j in range(updated_fixed_points.shape[0]):
                    updated_fixed_points[j,:] = total_field_xfrm.apply_to_point(tuple(fixed_points[j,:]))

            error_values.append(np.mean(np.sqrt(np.sum(np.square(updated_fixed_points - moving_points), axis=1, keepdims=True))))
            convergence_value = ants.convergence_monitoring(error_values)
            if verbose:
                end_time = time.time()
                diff_time = end_time - start_time
                print("Composition " + str(i) + ": error = " + str(error_values[-1]) +
                      " (convergence = " + str(convergence_value) + ", elapsed time = " + str(diff_time) + ")")
            if not convergence_value is None and convergence_value <= convergence_threshold:
                break

        if verbose:
            end_total_time = time.time()
            diff_total_time = end_total_time - start_total_time
            print("Total elapsed time = " + str(diff_total_time) + ".")

        return(total_field_xfrm)

    elif transform_type == "syn":

        if verbose:
            start_total_time = time.time()

        updated_fixed_points = np.empty_like(fixed_points)
        updated_fixed_points[:] = fixed_points
        updated_moving_points = np.empty_like(moving_points)
        updated_moving_points[:] = moving_points

        total_field_fixed_to_middle = create_zero_displacement_field(domain_image)
        total_inverse_field_fixed_to_middle = create_zero_displacement_field(domain_image)

        total_field_moving_to_middle = create_zero_displacement_field(domain_image)
        total_inverse_field_moving_to_middle = create_zero_displacement_field(domain_image)

        error_values = []
        for i in range(number_of_compositions):

            if verbose:
                start_time = time.time()

            update_field_fixed_to_middle = ants.fit_bspline_displacement_field(
              displacement_origins=updated_fixed_points,
              displacements=updated_moving_points - updated_fixed_points,
              displacement_weights=displacement_weights,
              origin=domain_image.origin,
              spacing=domain_image.spacing,
              size=domain_image.shape,
              direction=domain_image.direction,
              number_of_fitting_levels=number_of_fitting_levels,
              mesh_size=mesh_size,
              spline_order=spline_order,
              enforce_stationary_boundary=True,
              rasterize_points=rasterize_points
            )

            update_field_moving_to_middle = ants.fit_bspline_displacement_field(
              displacement_origins=updated_moving_points,
              displacements=updated_fixed_points - updated_moving_points,
              displacement_weights=displacement_weights,
              origin=domain_image.origin,
              spacing=domain_image.spacing,
              size=domain_image.shape,
              direction=domain_image.direction,
              number_of_fitting_levels=number_of_fitting_levels,
              mesh_size=mesh_size,
              spline_order=spline_order,
              enforce_stationary_boundary=True,
              rasterize_points=rasterize_points
            )

            update_field_fixed_to_middle = update_field_fixed_to_middle * composition_step_size
            update_field_moving_to_middle = update_field_moving_to_middle * composition_step_size
            if sigma > 0:
                update_field_fixed_to_middle = ants.smooth_image(update_field_fixed_to_middle, sigma)
                update_field_moving_to_middle = ants.smooth_image(update_field_moving_to_middle, sigma)

            # Add the update field to both forward displacement fields.

            total_field_fixed_to_middle = ants.compose_displacement_fields(update_field_fixed_to_middle, total_field_fixed_to_middle)
            total_field_moving_to_middle = ants.compose_displacement_fields(update_field_moving_to_middle, total_field_moving_to_middle)

            # Iteratively estimate the inverse fields.

            total_inverse_field_fixed_to_middle = ants.invert_displacement_field(total_field_fixed_to_middle, total_inverse_field_fixed_to_middle)
            total_inverse_field_moving_to_middle = ants.invert_displacement_field(total_field_moving_to_middle, total_inverse_field_moving_to_middle)

            total_field_fixed_to_middle = ants.invert_displacement_field(total_inverse_field_fixed_to_middle, total_field_fixed_to_middle)
            total_field_moving_to_middle = ants.invert_displacement_field(total_inverse_field_moving_to_middle, total_field_moving_to_middle)

            total_field_fixed_to_middle_xfrm = ants.transform_from_displacement_field(total_field_fixed_to_middle)
            total_field_moving_to_middle_xfrm = ants.transform_from_displacement_field(total_field_moving_to_middle)

            total_inverse_field_fixed_to_middle_xfrm = ants.transform_from_displacement_field(total_inverse_field_fixed_to_middle)
            total_inverse_field_moving_to_middle_xfrm = ants.transform_from_displacement_field(total_inverse_field_moving_to_middle)

            if i < number_of_compositions - 1:
                for j in range(updated_fixed_points.shape[0]):
                    updated_fixed_points[j,:] = total_field_fixed_to_middle_xfrm.apply_to_point(tuple(fixed_points[j,:]))
                    updated_moving_points[j,:] = total_field_moving_to_middle_xfrm.apply_to_point(tuple(moving_points[j,:]))

            error_values.append(np.mean(np.sqrt(np.sum(np.square(updated_fixed_points - updated_moving_points), axis=1, keepdims=True))))
            convergence_value = ants.convergence_monitoring(error_values)
            if verbose:
                end_time = time.time()
                diff_time = end_time - start_time
                print("Composition " + str(i) + ": error = " + str(error_values[-1]) +
                      " (convergence = " + str(convergence_value) + ", elapsed time = " + str(diff_time) + ")")
            if not convergence_value is None and convergence_value <= convergence_threshold:
                break

        total_forward_field = ants.compose_displacement_fields(total_inverse_field_moving_to_middle, total_field_fixed_to_middle)
        total_forward_xfrm = ants.transform_from_displacement_field(total_forward_field)
        total_inverse_field = ants.compose_displacement_fields(total_inverse_field_fixed_to_middle, total_field_moving_to_middle)
        total_inverse_xfrm = ants.transform_from_displacement_field(total_inverse_field)

        if verbose:
            end_total_time = time.time()
            diff_total_time = end_total_time - start_total_time
            print("Total elapsed time = " + str(diff_total_time) + ".")

        return_dict = {'forward_transform' : total_forward_xfrm,
                       'inverse_transform' : total_inverse_xfrm,
                       'fixed_to_middle_transform' : total_field_fixed_to_middle_xfrm,
                       'middle_to_fixed_transform' : total_inverse_field_fixed_to_middle_xfrm,
                       'moving_to_middle_transform' : total_field_moving_to_middle_xfrm,
                       'middle_to_moving_transform' : total_inverse_field_moving_to_middle_xfrm
                       }
        return(return_dict)

    elif transform_type == "tv" or transform_type == "time-varying":

        if verbose:
            start_total_time = time.time()

        updated_fixed_points = np.empty_like(fixed_points)
        updated_fixed_points[:] = fixed_points
        updated_moving_points = np.empty_like(moving_points)
        updated_moving_points[:] = moving_points

        velocity_field = create_zero_velocity_field(domain_image, number_of_time_steps)
        velocity_field_array = velocity_field.numpy()

        last_update_derivative_field = create_zero_velocity_field(domain_image, number_of_time_steps)
        last_update_derivative_field_array = last_update_derivative_field.numpy()

        error_values = []
        for i in range(number_of_compositions):

            if verbose:
                start_time = time.time()

            update_derivative_field = create_zero_velocity_field(domain_image, number_of_time_steps)
            update_derivative_field_array = update_derivative_field.numpy()

            average_error = 0.0
            for n in range(number_of_time_steps):

                t = n / (number_of_time_steps - 1.0)

                if n > 0:
                    integrated_forward_field = ants.integrate_velocity_field(velocity_field, 0.0, t, number_of_integration_steps)
                    integrated_forward_field_xfrm = ants.transform_from_displacement_field(integrated_forward_field)
                    for j in range(updated_fixed_points.shape[0]):
                        updated_fixed_points[j,:] = integrated_forward_field_xfrm.apply_to_point(tuple(fixed_points[j,:]))
                else:
                    updated_fixed_points[:] = fixed_points

                if n < number_of_time_steps - 1:
                    integrated_inverse_field = ants.integrate_velocity_field(velocity_field, 1.0, t, number_of_integration_steps)
                    integrated_inverse_field_xfrm = ants.transform_from_displacement_field(integrated_inverse_field)
                    for j in range(updated_moving_points.shape[0]):
                        updated_moving_points[j,:] = integrated_inverse_field_xfrm.apply_to_point(tuple(moving_points[j,:]))
                else:
                    updated_moving_points[:] = moving_points

                update_derivative_field_at_timepoint = ants.fit_bspline_displacement_field(
                  displacement_origins=updated_fixed_points,
                  displacements=updated_moving_points - updated_fixed_points,
                  displacement_weights=displacement_weights,
                  origin=domain_image.origin,
                  spacing=domain_image.spacing,
                  size=domain_image.shape,
                  direction=domain_image.direction,
                  number_of_fitting_levels=number_of_fitting_levels,
                  mesh_size=mesh_size,
                  spline_order=spline_order,
                  enforce_stationary_boundary=True,
                  rasterize_points=rasterize_points
                  )

                if sigma > 0:
                    update_derivative_field_at_timepoint = ants.smooth_image(update_derivative_field_at_timepoint, sigma)

                update_derivative_field_at_timepoint_array = update_derivative_field_at_timepoint.numpy()
                grad_norms = np.sqrt(np.sum(np.square(update_derivative_field_at_timepoint_array), axis=-1, keepdims=False))
                max_norm = np.amax(grad_norms)
                median_norm = np.median(grad_norms)
                if verbose:
                    print("  integration point " + str(t) + ": max_norm = " + str(max_norm) + ", median_norm = " + str(median_norm))
                update_derivative_field_at_timepoint_array /= max_norm
                if domain_image.dimension == 2:
                    update_derivative_field_array[:,:,n,:] = update_derivative_field_at_timepoint_array
                elif domain_image.dimension == 3:
                    update_derivative_field_array[:,:,:,n,:] = update_derivative_field_at_timepoint_array

                rmse = np.mean(np.sqrt(np.sum(np.square(updated_moving_points - updated_fixed_points), axis=1, keepdims=True)))
                average_error = (average_error * n + rmse) / (n + 1)

            update_derivative_field_array = (update_derivative_field_array + last_update_derivative_field_array) * 0.5
            last_update_derivative_field_array = np.empty_like(update_derivative_field_array)
            last_update_derivative_field_array[:] = update_derivative_field_array

            velocity_field_array = velocity_field_array + update_derivative_field_array * composition_step_size
            velocity_field = ants.from_numpy(velocity_field_array, origin=velocity_field.origin,
                                             spacing=velocity_field.spacing, direction=velocity_field.direction,
                                             has_components=True)

            error_values.append(average_error)
            convergence_value = ants.convergence_monitoring(error_values)
            if verbose:
                end_time = time.time()
                diff_time = end_time - start_time
                print("Composition " + str(i) + ": error = " + str(error_values[-1]) +
                      " (convergence = " + str(convergence_value) + ", elapsed time = " + str(diff_time) + ")")
            if not convergence_value is None and convergence_value <= convergence_threshold:
                break

        forward_xfrm = ants.transform_from_displacement_field(ants.integrate_velocity_field(velocity_field, 0.0, 1.0, number_of_integration_steps))
        inverse_xfrm = ants.transform_from_displacement_field(ants.integrate_velocity_field(velocity_field, 1.0, 0.0, number_of_integration_steps))

        if verbose:
            end_total_time = time.time()
            diff_total_time = end_total_time - start_total_time
            print("Total elapsed time = " + str(diff_total_time) + ".")

        return_dict = {'forward_transform': forward_xfrm,
                       'inverse_transform': inverse_xfrm,
                       'velocity_field': velocity_field}
        return(return_dict)

    else:
        raise ValueError("Unrecognized transform_type.")


def fit_time_varying_transform_to_point_sets(point_sets,
                                             time_points=None,
                                             initial_velocity_field=None,
                                             number_of_time_steps=None,
                                             domain_image=None,
                                             number_of_fitting_levels=4,
                                             mesh_size=1,
                                             spline_order=3,
                                             displacement_weights=None,
                                             number_of_compositions=10,
                                             composition_step_size=0.5,
                                             number_of_integration_steps=100,
                                             sigma=0.0,
                                             convergence_threshold=1e-6,
                                             rasterize_points=False,
                                             verbose=False
                                            ):
    """

    Estimate a time-varying transform from corresponding point sets (> 2).

    ANTsR function: fitTimeVaryingTransformToPointSets

    Arguments
    ---------
    point_sets : list of arrays
        Corresponding points across sets specified in physical space as a n x d matrix where n
        is the number of points and d is the dimensionality.

    time_points : array of ordered scalars between 0 and 1
        Set of scalar values, one for each point-set, designating its time position in the velocity
        flow.  If not set, it defaults to equal spacing between 0 and 1.

    initial_velocity_field : initial ANTs velocity field
        Optional velocity field for initializing optimization.  Overrides the number of integration
        points.

    number_of_time_steps : integer
        Time-varying velocity field parameter.  Needs to be equal to or greater than the number of
        point sets.  If not specified, it defaults to the number of point sets.

    domain_image : ANTs image
        Defines physical domain of the nonlinear transform.  Must be defined.

    number_of_fitting_levels : integer
        Integer specifying the number of fitting levels for the B-spline interpolation of the
        displacement field.

    mesh_size : integer or array
        Defines the mesh size at the initial fitting level for the B-spline interpolation of the
        displacement field..

    spline_order : integer
        Spline order of the B-spline displacement field.

    displacement_weights : array
        Defines the individual weighting of the corresponding scattered data value.  Default = NULL
        meaning all displacements are weighted the same.

    number_of_compositions : integer
        Total number of compositions.

    composition_step_size : scalar
        Scalar multiplication factor of the weighting of the update field.

    number_of_integration_steps : scalar
        Number of steps used for integrating the velocity field.

    sigma : scalar
        Gaussian smoothing standard deviation of the update field (in mm).

    convergence_threshold : scalar
        Composition-based convergence parameter using a window size of 10 values.

    rasterize_points : boolean
        Use nearest neighbor rasterization of points for estimating the update field (potential
        speed-up).  Default = False.

    verbose : bool
        Print progress to the screen.

    Returns
    -------

    ANTs transform

    Example
    -------
    >>> import ants
    >>> import numpy as np
    """

    def create_zero_velocity_field(domain_image, number_of_time_points=2):
         field_array = np.zeros((*domain_image.shape, number_of_time_points, domain_image.dimension))
         origin = (*domain_image.origin, 0.0)
         spacing = (*domain_image.spacing, 1.0)
         direction = np.eye(domain_image.dimension + 1)
         direction[0:domain_image.dimension,0:domain_image.dimension] = domain_image.direction
         field = ants.from_numpy(field_array, origin=origin, spacing=spacing, direction=direction,
                 has_components=True)
         return(field)

    if not isinstance(point_sets, list):
        raise ValueError("point_sets should be a list of corresponding point sets.")

    number_of_point_sets = len(point_sets)

    if time_points is not None and len(time_points) != number_of_point_sets:
        raise ValueError("The number of time points should be the same as the number of point sets.")

    if time_points is None:
        time_points = np.linspace(0.0, 1.0, number_of_point_sets)
    time_points = np.array(time_points)

    if np.any(time_points < 0.0) or np.any(time_points > 1.0):
        raise ValueError("time point values should be between 0 and 1.")

    if number_of_point_sets < 3:
        raise ValueError("Expecting three or greater point sets.")

    if domain_image is None:
        raise ValueError("Domain image needs to be specified.")

    number_of_points = point_sets[0].shape[0]
    dimensionality = point_sets[0].shape[1]
    for i in range(1, number_of_point_sets):
        if point_sets[i].shape[0] != number_of_points:
            raise ValueError("Point sets should match in terms of the number of points.")
        if point_sets[i].shape[1] != dimensionality:
            raise ValueError("Point sets should match in terms of dimensionality.")

    if verbose:
        start_total_time = time.time()

    updated_fixed_points = np.zeros(point_sets[0].shape)
    updated_moving_points = np.zeros(point_sets[0].shape)

    velocity_field = None
    if initial_velocity_field is None:
        if number_of_time_steps is None:
            number_of_time_steps = len(time_points)
        if number_of_time_steps < number_of_point_sets:
            raise ValueError("The number of integration points should be at least as great as the number of point sets.")
        velocity_field = create_zero_velocity_field(domain_image, number_of_time_steps)
    else:
        velocity_field = ants.image_clone(initial_velocity_field)
        number_of_time_steps = initial_velocity_field.shape[-1]
    velocity_field_array = velocity_field.numpy()

    last_update_derivative_field = create_zero_velocity_field(domain_image, number_of_time_steps)
    last_update_derivative_field_array = last_update_derivative_field.numpy()

    error_values = []
    for i in range(number_of_compositions):

        if verbose:
            start_time = time.time()

        update_derivative_field = create_zero_velocity_field(domain_image, number_of_time_steps)
        update_derivative_field_array = update_derivative_field.numpy()

        average_error = 0.0
        for n in range(number_of_time_steps):

            t = n / (number_of_time_steps - 1.0)

            t_index = 0
            for j in range(1, number_of_point_sets):
                if time_points[j-1] <= t and time_points[j] >= t:
                    t_index = j
                    break

            if n > 0 and n < number_of_time_steps - 1 and time_points[t_index-1] == t:
                updated_fixed_points[:] = point_sets[t_index-1]
                integrated_inverse_field = ants.integrate_velocity_field(velocity_field, time_points[t_index], t, number_of_integration_steps)
                integrated_inverse_field_xfrm = ants.transform_from_displacement_field(integrated_inverse_field)
                for j in range(updated_moving_points.shape[0]):
                    updated_moving_points[j,:] = integrated_inverse_field_xfrm.apply_to_point(tuple(point_sets[t_index][j,:]))

                update_derivative_field_at_timepoint_forward = ants.fit_bspline_displacement_field(
                  displacement_origins=updated_fixed_points,
                  displacements=updated_moving_points - updated_fixed_points,
                  displacement_weights=displacement_weights,
                  origin=domain_image.origin,
                  spacing=domain_image.spacing,
                  size=domain_image.shape,
                  direction=domain_image.direction,
                  number_of_fitting_levels=number_of_fitting_levels,
                  mesh_size=mesh_size,
                  spline_order=spline_order,
                  enforce_stationary_boundary=True,
                  rasterize_points=rasterize_points
                  )

                updated_moving_points[:] = point_sets[t_index-1]
                integrated_forward_field = ants.integrate_velocity_field(velocity_field, time_points[t_index-2], t, number_of_integration_steps)
                integrated_forward_field_xfrm = ants.transform_from_displacement_field(integrated_forward_field)
                for j in range(updated_fixed_points.shape[0]):
                    updated_fixed_points[j,:] = integrated_forward_field_xfrm.apply_to_point(tuple(point_sets[t_index-2][j,:]))

                update_derivative_field_at_timepoint_back = ants.fit_bspline_displacement_field(
                  displacement_origins=updated_fixed_points,
                  displacements=updated_moving_points - updated_fixed_points,
                  displacement_weights=displacement_weights,
                  origin=domain_image.origin,
                  spacing=domain_image.spacing,
                  size=domain_image.shape,
                  direction=domain_image.direction,
                  number_of_fitting_levels=number_of_fitting_levels,
                  mesh_size=mesh_size,
                  spline_order=spline_order,
                  enforce_stationary_boundary=True,
                  rasterize_points=rasterize_points
                  )

                update_derivative_field_at_timepoint = (update_derivative_field_at_timepoint_forward +
                                                        update_derivative_field_at_timepoint_back) / 2.0

            else:
                if t == 0.0 and time_points[t_index-1] == 0.0:
                    updated_fixed_points[:] = point_sets[0]
                else:
                    integrated_forward_field = ants.integrate_velocity_field(velocity_field, time_points[t_index-1], t, number_of_integration_steps)
                    integrated_forward_field_xfrm = ants.transform_from_displacement_field(integrated_forward_field)
                    for j in range(updated_fixed_points.shape[0]):
                        updated_fixed_points[j,:] = integrated_forward_field_xfrm.apply_to_point(tuple(point_sets[t_index-1][j,:]))

                if t == 1.0 and time_points[t_index] == 1.0:
                    updated_moving_points[:] = point_sets[-1]
                else:
                    integrated_inverse_field = ants.integrate_velocity_field(velocity_field, time_points[t_index], t, number_of_integration_steps)
                    integrated_inverse_field_xfrm = ants.transform_from_displacement_field(integrated_inverse_field)
                    for j in range(updated_moving_points.shape[0]):
                        updated_moving_points[j,:] = integrated_inverse_field_xfrm.apply_to_point(tuple(point_sets[t_index][j,:]))

                update_derivative_field_at_timepoint = ants.fit_bspline_displacement_field(
                  displacement_origins=updated_fixed_points,
                  displacements=updated_moving_points - updated_fixed_points,
                  displacement_weights=displacement_weights,
                  origin=domain_image.origin,
                  spacing=domain_image.spacing,
                  size=domain_image.shape,
                  direction=domain_image.direction,
                  number_of_fitting_levels=number_of_fitting_levels,
                  mesh_size=mesh_size,
                  spline_order=spline_order,
                  enforce_stationary_boundary=True,
                  rasterize_points=rasterize_points
                  )

            if sigma > 0:
                update_derivative_field_at_timepoint = ants.smooth_image(update_derivative_field_at_timepoint, sigma)

            update_derivative_field_at_timepoint_array = update_derivative_field_at_timepoint.numpy()
            grad_norms = np.sqrt(np.sum(np.square(update_derivative_field_at_timepoint_array), axis=-1, keepdims=False))
            max_norm = np.amax(grad_norms)
            median_norm = np.median(grad_norms)
            if verbose:
                print("  integration point " + str(t) + ": max_norm = " + str(max_norm) + ", median_norm = " + str(median_norm))
            update_derivative_field_at_timepoint_array /= max_norm
            if domain_image.dimension == 2:
                update_derivative_field_array[:,:,n,:] = update_derivative_field_at_timepoint_array
            elif domain_image.dimension == 3:
                update_derivative_field_array[:,:,:,n,:] = update_derivative_field_at_timepoint_array

            rmse = np.mean(np.sqrt(np.sum(np.square(updated_moving_points - updated_fixed_points), axis=1, keepdims=True)))
            average_error = (average_error * n + rmse) / (n + 1)

        update_derivative_field_array = (update_derivative_field_array + last_update_derivative_field_array) * 0.5
        last_update_derivative_field_array = np.empty_like(update_derivative_field_array)
        last_update_derivative_field_array[:] = update_derivative_field_array

        velocity_field_array += (update_derivative_field_array * composition_step_size)
        velocity_field = ants.from_numpy(velocity_field_array, origin=velocity_field.origin,
                                         spacing=velocity_field.spacing, direction=velocity_field.direction,
                                         has_components=True)

        error_values.append(average_error)
        convergence_value = ants.convergence_monitoring(error_values)
        if verbose:
            end_time = time.time()
            diff_time = end_time - start_time
            print("Composition " + str(i) + ": error = " + str(error_values[-1]) +
                  " (convergence = " + str(convergence_value) + ", elapsed time = " + str(diff_time) + ")")
        if not convergence_value is None and convergence_value <= convergence_threshold:
            break

    forward_xfrm = ants.transform_from_displacement_field(ants.integrate_velocity_field(velocity_field, 0.0, 1.0, number_of_integration_steps))
    inverse_xfrm = ants.transform_from_displacement_field(ants.integrate_velocity_field(velocity_field, 1.0, 0.0, number_of_integration_steps))

    if verbose:
        end_total_time = time.time()
        diff_total_time = end_total_time - start_total_time
        print("Total elapsed time = " + str(diff_total_time) + ".")

    return_dict = {'forward_transform': forward_xfrm,
                   'inverse_transform': inverse_xfrm,
                   'velocity_field': velocity_field}
    return(return_dict)








__all__ = ['ilr',
           'rank_intensity',
           'quantile',
           'regress_poly',
           'regress_components',
           'get_average_of_timeseries',
           'compcor',
           'bandpass_filter_matrix' ]

import numpy as np
from numpy.polynomial import Legendre
from scipy import linalg
from scipy.stats import pearsonr
from scipy.stats import rankdata
import pandas as pd
from pandas import DataFrame
import statsmodels.api as sm
import statsmodels.formula.api as smf

import ants
from ants.decorators import image_method

def rank_intensity( x, mask=None, get_mask=True, method='max',  ):
    """
    Rank transform the intensity of the input image with or without masking.
    Intensities will transform from [0,1,2,55] to [0,1,2,3] so this may not be
    appropriate for quantitative images - however, you never know.  rank
    transformations generally improve robustness so it is an empirical question
    that should be evaluated.

    Arguments
    ---------

    x : ANTsImage
        input image

    mask : ANTsImage
        optional mask

    get_mask: boolean
        will estimate a mask when none provided

    method : a scipy rank method (max,min,average,dense)


    return: transformed image

    Example
    -------
    >>> img = ants.image_read(ants.get_data('r16'))
    >>> ants.rank_intensity(img)
    """
    if mask is not None:
        fir = rankdata( (x*mask).numpy(), method=method )
    elif mask is None and get_mask == True:
        mask = ants.get_mask( x )
        fir = rankdata( (x*mask).numpy(), method=method )
    else:
        fir = rankdata( x.numpy(), method=method )
    fir = fir - 1
    fir = fir.reshape( x.shape )
    rimg = ants.from_numpy( fir.astype(float)  )
    rimg = ants.iMath(rimg,"Normalize")
    ants.copy_image_info( x, rimg )
    if mask is not None:
        rimg = rimg * mask
    return( rimg )


def ilr( data_frame, voxmats, ilr_formula, verbose = False ):
    """
    Image-based linear regression.

    This function simplifies calculating p-values from linear models
    in which there is a similar formula that is applied many times
    with a change in image-based predictors.  Image-based variables
    are stored in the input matrix list. They should be named
    consistently in the input formula and in the image list.  If they
    are not, an error will be thrown.  All input matrices should have
    the same number of rows and columns.

    This function takes advantage of statsmodels R-style formulas.

    ANTsR function: `ilr`

    Arguments
    ---------

    data_frame: This data frame contains all relevant predictors except for
        the matrices associated with the image variables.  One should convert
        any categorical predictors ahead of time using `pd.get_dummies`.

    voxmats: The named list of matrices that contains the changing
        predictors.

    ilr_formula: This is a character string that defines a valid regression
        formula in the R-style.

    verbose:  will print a little bit of diagnostic information that allows
        a degree of model checking

    Returns
    -------

    A list of different matrices that contain names derived from the
    formula and the coefficients of the regression model.  The size of
    the output values ( p-values, t-values, parameter values ) will match
    the input matrix and, as such, can be converted to an image via `make_image`

    Example
    -------

    >>> nsub = 20
    >>> mu, sigma = 0, 1
    >>> outcome = np.random.normal( mu, sigma, nsub )
    >>> covar = np.random.normal( mu, sigma, nsub )
    >>> mat = np.random.normal( mu, sigma, (nsub, 500 ) )
    >>> mat2 = np.random.normal( mu, sigma, (nsub, 500 ) )
    >>> data = {'covar':covar,'outcome':outcome}
    >>> df = pd.DataFrame( data )
    >>> vlist = { "mat1": mat, "mat2": mat2 }
    >>> myform = " outcome ~ covar * mat1 "
    >>> result = ants.ilr( df, vlist, myform)
    >>> myform = " mat2 ~ covar + mat1 "
    >>> result = ants.ilr( df, vlist, myform)

    """

    nvoxmats = len( voxmats )
    if nvoxmats < 1 :
        raise ValueError('Pass at least one matrix to voxmats list')
    keylist = list(voxmats.keys())
    firstmat = keylist[0]
    voxshape = voxmats[firstmat].shape
    nvox = voxshape[1]
    nmats = len( keylist )
    for k in keylist:
        if voxmats[firstmat].shape != voxmats[k].shape:
            raise ValueError('Matrices must have same number of rows (samples)')

    # test voxel
    vox = 0
    nrows = data_frame.shape[0]
    data_frame_vox = data_frame.copy()
    for k in range( nmats ):
        data = {keylist[k]: np.random.normal(0,1,nrows) }
        temp = pd.DataFrame( data )
        data_frame_vox = pd.concat([data_frame_vox.reset_index(drop=True),temp], axis=1 )
    mod = smf.ols(formula=ilr_formula, data=data_frame_vox )
    res = mod.fit()
    modelNames = res.model.exog_names
    if verbose:
        print( data_frame_vox )
        print(res.summary())
    nOutcomes = len( modelNames )
    tValsOut = list()
    pValsOut = list()
    bValsOut = list()
    for k in range( len( modelNames ) ):
        bValsOut.append( np.zeros( nvox ) )
        pValsOut.append( np.zeros( nvox ) )
        tValsOut.append( np.zeros( nvox ) )

    data_frame_vox = data_frame.copy()
    for v in range( nmats ):
        data = {keylist[v]: voxmats[keylist[v]][:,k] }
        temp = pd.DataFrame( data )
        data_frame_vox = pd.concat([data_frame_vox.reset_index(drop=True),temp], axis=1 )
    for k in range( nvox ):
        # first get the correct data frame
        for v in range( nmats ):
            data_frame_vox[ keylist[v] ] = voxmats[keylist[v]][:,k]
        # then get the local model results
        mod = smf.ols(formula=ilr_formula, data=data_frame_vox )
        res = mod.fit()
        tvals = res.tvalues
        pvals = res.pvalues
        bvals = res.params
        for v in range( len( modelNames ) ):
            bValsOut[v][k] = bvals.iloc[v]
            pValsOut[v][k] = pvals.iloc[v]
            tValsOut[v][k] = tvals.iloc[v]

    bValsOutDict = { }
    tValsOutDict = { }
    pValsOutDict = { }
    for v in range( len( modelNames ) ):
        bValsOutDict[ 'coef_' + modelNames[v] ] = bValsOut[v]
        tValsOutDict[ 'tval_' + modelNames[v] ] = tValsOut[v]
        pValsOutDict[ 'pval_' + modelNames[v] ] = pValsOut[v]

    return {
        'modelNames': modelNames,
        'coefficientValues': bValsOutDict,
        'pValues': pValsOutDict,
        'tValues': tValsOutDict }


@image_method
def quantile(image, q, nonzero=True):
    """
    Get the quantile values from an ANTsImage
    
    Examples
    --------
    >>> img = ants.image_read(ants.get_data('r16'))
    >>> ants.quantile(img, 0.5)
    >>> ants.quantile(img, (0.5, 0.75))
    """
    img_arr = image.numpy()
    if isinstance(q, (list,tuple)):
        q = [qq*100. if qq <= 1. else qq for qq in q]
        if nonzero:
            img_arr = img_arr[img_arr>0]
        vals = [np.percentile(img_arr, qq) for qq in q]
        return tuple(vals)
    elif isinstance(q, (float,int)):
        if q <= 1.:
            q = q*100.
        if nonzero:
            img_arr = img_arr[img_arr>0]
        return np.percentile(img_arr[img_arr>0], q)
    else:
        raise ValueError('q argument must be list/tuple or float/int')


def regress_poly(degree, data, remove_mean=True, axis=-1):
    """
    Returns data with degree polynomial regressed out.
    :param bool remove_mean: whether or not demean data (i.e. degree 0),
    :param int axis: numpy array axes along which regression is performed
    """
    timepoints = data.shape[0]
    # Generate design matrix
    X = np.ones((timepoints, 1))  # quick way to calc degree 0
    for i in range(degree):
        polynomial_func = Legendre.basis(i + 1)
        value_array = np.linspace(-1, 1, timepoints)
        X = np.hstack((X, polynomial_func(value_array)[:, np.newaxis]))
    non_constant_regressors = X[:, :-1] if X.shape[1] > 1 else np.array([])
    betas = np.linalg.pinv(X).dot(data)
    if remove_mean:
        datahat = X.dot(betas)
    else:  # disregard the first layer of X, which is degree 0
        datahat = X[:, 1:].dot(betas[1:, ...])
    regressed_data = data - datahat
    return regressed_data, non_constant_regressors

def regress_components( data, components, remove_mean=True ):
    """
    Returns data with components regressed out.
    :param bool remove_mean: whether or not demean data (i.e. degree 0),
    :param int axis: numpy array axes along which regression is performed
    """
    timepoints = data.shape[0]
    betas = np.linalg.pinv(components).dot(data)
    if remove_mean:
        datahat = components.dot(betas)
    else:  # disregard the first layer of X, which is degree 0
        datahat = components[:, 1:].dot(betas[1:, ...])
    regressed_data = data - datahat
    return regressed_data


def get_average_of_timeseries( image, idx=None ):
    """Average the timeseries into a dimension-1 image.
    image: input time series image
    idx:  indices over which to average
    """
    imagedim = image.dimension
    if idx is None:
        idx = range( image.shape[ imagedim - 1 ] )
    i0 = ants.slice_image( image, axis=image.dimension-1, idx=idx[0] ) * 0
    wt = 1.0 / len( idx )
    for k in idx:
        i0 = i0 + ants.slice_image( image, axis=image.dimension-1, idx=k ) * wt
    return( i0 )

def bandpass_filter_matrix( matrix,
    tr=1, lowf=0.01, highf=0.1, order = 3):
    """
    Bandpass filter the input time series image

    ANTsR function: `frequencyFilterfMRI`

    Arguments
    ---------

    image: input time series image

    tr:    sampling time interval (inverse of sampling rate)

    lowf:  low frequency cutoff

    highf: high frequency cutoff

    order: order of the butterworth filter run using `filtfilt`

    Returns
    -------
    filtered matrix

    Example
    -------

    >>> import numpy as np
    >>> import ants
    >>> import matplotlib.pyplot as plt
    >>> brainSignal = np.random.randn( 400, 1000 )
    >>> tr = 1
    >>> filtered = ants.bandpass_filter_matrix( brainSignal, tr = tr )
    >>> nsamples = brainSignal.shape[0]
    >>> t = np.linspace(0, tr*nsamples, nsamples, endpoint=False)
    >>> k = 20
    >>> plt.plot(t, brainSignal[:,k], label='Noisy signal')
    >>> plt.plot(t, filtered[:,k], label='Filtered signal')
    >>> plt.xlabel('time (seconds)')
    >>> plt.grid(True)
    >>> plt.axis('tight')
    >>> plt.legend(loc='upper left')
    >>> plt.show()
    """
    from scipy.signal import butter, filtfilt

    def butter_bandpass(lowcut, highcut, fs, order ):
        nyq = 0.5 * fs
        low = lowcut / nyq
        high = highcut / nyq
        b, a = butter(order, [low, high], btype='band')
        return b, a

    def butter_bandpass_filter(data, lowcut, highcut, fs, order ):
        b, a = butter_bandpass(lowcut, highcut, fs, order=order)
        y = filtfilt(b, a, data)
        return y

    fs = 1/tr   # sampling rate based on tr
    nsamples = matrix.shape[0]
    ncolumns = matrix.shape[1]
    matrixOut = matrix.copy()
    for k in range( ncolumns ):
        matrixOut[:,k] = butter_bandpass_filter(
            matrix[:,k], lowf, highf, fs, order=order )
    return matrixOut

def clean_data(arr, standardize=True):
    """
    Remove columns from a NumPy array that have no variation or contain NA/Inf values.
    Optionally standardize the remaining data.

    :param arr: NumPy array to be cleaned.
    :param standardize: Boolean, if True standardize the data.
    :return: Cleaned (and optionally standardized) NumPy array.
    """
    valid_columns = []

    for i in range(arr.shape[1]):
        column = arr[:, i]
        if np.any(column != column[0]) and not np.any(np.isnan(column)) and not np.any(np.isinf(column)):
            valid_columns.append(i)

    cleaned_data = arr[:, valid_columns]

    if standardize:
        mean = np.mean(cleaned_data, axis=0)
        std_dev = np.std(cleaned_data, axis=0)
        # Avoid division by zero in case of zero standard deviation
        std_dev[std_dev == 0] = 1
        cleaned_data = (cleaned_data - mean) / std_dev

    return cleaned_data

def compcor( boldImage, ncompcor=4, quantile=0.975, mask=None, filter_type=False, degree=2 ):
    """
    Compute noise components from the input image

    ANTsR function: `compcor`

    this is adapted from nipy code https://github.com/nipy/nipype/blob/e29ac95fc0fc00fedbcaa0adaf29d5878408ca7c/nipype/algorithms/confounds.py

    Arguments
    ---------

    boldImage: input time series image

    ncompcor:  number of noise components to return

    quantile:  quantile defining high-variance

    mask:      mask defining brain or specific tissues

    filter_type: type off filter to apply to time series before computing
                 noise components.

        'polynomial' - Legendre polynomial basis
        False - None (mean-removal only)

    degree: order of polynomial used to remove trends from the timeseries

    Returns
    -------
    dictionary containing:

        components: a numpy array

        basis: a numpy array containing the (non-constant) filter regressors

    Example
    -------
    >>> cc = ants.compcor( ants.image_read(ants.get_ants_data("ch2")) )

    """

    def compute_tSTD(M, quantile, x=0, axis=0):
        stdM = np.std(M, axis=axis)
        # set bad values to x
        stdM[stdM == 0] = x
        stdM[np.isnan(stdM)] = x
        tt = round( quantile*100 )
        threshold_std = np.percentile( stdM, tt )
    #    threshold_std = quantile( stdM, quantile )
        return { 'tSTD': stdM, 'threshold_std': threshold_std}
    if mask is None:
        temp = ants.slice_image( boldImage, axis=boldImage.dimension-1, idx=0 )
        mask = ants.get_mask( temp )
    imagematrix = ants.timeseries_to_matrix( boldImage, mask )
    temp = compute_tSTD( imagematrix, quantile, 0 )
    tsnrmask = ants.make_image( mask,  temp['tSTD'] )
    tsnrmask = ants.threshold_image( tsnrmask, temp['threshold_std'], temp['tSTD'].max() )
    M = ants.timeseries_to_matrix( boldImage, tsnrmask )
    components = None
    basis = np.array([])
    if filter_type in ('polynomial', False):
        M, basis = regress_poly(degree, M)
#    M = M / compute_tSTD(M, 1.)['tSTD']
    # "The covariance matrix C = MMT was constructed and decomposed into its
    # principal components using a singular value decomposition."
    M = clean_data( M, standardize=True )
    u, _, _ = linalg.svd(M, full_matrices=False)
    if components is None:
        components = u[:, :ncompcor]
    else:
        components = np.hstack((components, u[:, :ncompcor]))
    if components is None and ncompcor > 0:
        raise ValueError('No components found')
    return { 'components': components, 'basis': basis }





__all__ = ['make_points_image']

import math
import numpy as np

import ants

def make_points_image(pts, target, radius=5):
    """
    Create label image from physical space points

    Creates spherical points in the coordinate space of the target image based
    on the n-dimensional matrix of points that the user supplies. The image
    defines the dimensionality of the data so if the input image is 3D then
    the input points should be 2D or 3D.

    ANTsR function: `makePointsImage`

    Arguments
    ---------
    pts : numpy.ndarray
        input points

    target : ANTsImage
        Image defining target space

    radius : integer
        radius for the points

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> import pandas as pd
    >>> mni = ants.image_read(ants.get_data('mni'))
    >>> powers_pts = pd.read_csv(ants.get_data('powers_mni_itk'))
    >>> powers_labels = ants.make_points_image(powers_pts.iloc[:,:3].values, mni, radius=3)
    """
    lblimg = target * 0
    npts = len(pts)
    dim = target.dimension
    if pts.shape[1] != dim:
        raise ValueError('points dimensionality should match that of images')

    for r in range(npts):
        pt = pts[r,:]
        idx = ants.transform_physical_point_to_index(target, pt.tolist() ).astype(int)
        in_image=True
        for kk in range(target.dimension):
            in_image = in_image and idx[kk] >= 0 and idx[kk] < target.shape[kk]
        if ( in_image == True ):
            if (dim == 3):
                lblimg[idx[0],idx[1],idx[2]] = r + 1
            elif (dim == 2):
                lblimg[idx[0],idx[1]] = r + 1
    return ants.morphology( lblimg, 'dilate', radius, 'grayscale' )



"""
Functions for plotting ants images
"""


__all__ = [
    "movie"
]

import fnmatch
import math
import os
import warnings

from matplotlib import gridspec
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import matplotlib.lines as mlines
import matplotlib.patches as patches
import matplotlib.mlab as mlab
import matplotlib.animation as animation
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


import numpy as np
from ants.decorators import image_method

@image_method
def movie(image, filename=None, writer=None, fps=30):
    """
    Create and save a movie - mp4, gif, etc - of the various
    2D slices of a 3D ants image

    Try this:
        conda install -c conda-forge ffmpeg

    Example
    -------
    >>> import ants
    >>> mni = ants.image_read(ants.get_data('mni'))
    >>> ants.movie(mni, filename='~/desktop/movie.mp4')
    """

    image = image.pad_image()
    img_arr = image.numpy()

    minidx = max(0, np.where(image > 0)[0][0] - 5)
    maxidx = max(image.shape[0], np.where(image > 0)[0][-1] + 5)

    # Creare your figure and axes
    fig, ax = plt.subplots(1)

    im = ax.imshow(
        img_arr[minidx, :, :],
        animated=True,
        cmap="Greys_r",
        vmin=image.quantile(0.05),
        vmax=image.quantile(0.95),
    )

    ax.axis("off")

    def init():
        fig.axes("off")
        return (im,)

    def updatefig(frame):
        im.set_array(img_arr[frame, :, :])
        return (im,)

    ani = animation.FuncAnimation(
        fig,
        updatefig,
        frames=np.arange(minidx, maxidx),
        # init_func=init,
        interval=50,
        blit=True,
    )

    if writer is None:
        writer = animation.FFMpegWriter(fps=fps)

    if filename is not None:
        filename = os.path.expanduser(filename)
        ani.save(filename, writer=writer)
    else:
        plt.show()



from .ants_image_io import (image_header_info,
                            image_clone,
                            image_read,
                            dicom_read,
                            image_write,
                            make_image,
                            from_numpy,
                            from_numpy_like,
                            new_image_like)
from .ants_image import (ANTsImage,
                         copy_image_info,
                         set_origin,
                         get_origin,
                         set_direction,
                         get_direction,
                         set_spacing,
                         get_spacing,
                         is_image,
                         from_pointer)
from .ants_metric_io import (new_ants_metric,
                             create_ants_metric,
                             supported_metrics)
from .ants_transform_io import (create_ants_transform,
                                new_ants_transform,
                                read_transform,
                                write_transform,
                                transform_from_displacement_field,
                                transform_to_displacement_field,
                                fsl2antstransform)
from .ants_transform import (ANTsTransform,
                             set_ants_transform_parameters,
                             get_ants_transform_parameters,
                             get_ants_transform_fixed_parameters,
                             set_ants_transform_fixed_parameters,
                             apply_ants_transform,
                             apply_ants_transform_to_point,
                             apply_ants_transform_to_vector,
                             apply_ants_transform_to_image,
                             invert_ants_transform,
                             compose_ants_transforms,
                             transform_index_to_physical_point,
                             transform_physical_point_to_index)


from .add_noise_to_image import add_noise_to_image
from .anti_alias import anti_alias
from .bias_correction import (n3_bias_field_correction,
                              n3_bias_field_correction2,
                              n4_bias_field_correction,
                              abp_n4)
from .crop_image import (crop_image,
                         crop_indices,
                         decrop_image)
from .denoise_image import denoise_image
from .get_mask import get_mask
from .image_type_cast import image_type_cast
from .hessian_objectness import hessian_objectness
from .histogram_equalize_image import histogram_equalize_image
from .histogram_match_image import histogram_match_image, histogram_match_image2
from .iMath import (iMath,
                    image_math,
                    multiply_images,
                    iMath_get_largest_component,
                    iMath_normalize,
                    iMath_truncate_intensity,
                    iMath_sharpen,
                    iMath_pad,
                    iMath_maurer_distance,
                    iMath_perona_malik,
                    iMath_grad,
                    iMath_laplacian,
                    iMath_canny,
                    iMath_histogram_equalization,
                    iMath_MD,
                    iMath_ME,
                    iMath_MO,
                    iMath_MC,
                    iMath_GD,
                    iMath_GE,
                    iMath_GO,
                    iMath_GC,
                    iMath_fill_holes,
                    iMath_get_largest_component,
                    iMath_normalize,
                    iMath_truncate_intensity,
                    iMath_sharpen,
                    iMath_propagate_labels_through_mask)
from .mask_image import mask_image
from .morphology import morphology
from .pad_image import pad_image
from .reflect_image import reflect_image
from .reorient_image import (get_orientation,
                             reorient_image2,
                             get_possible_orientations,
                             get_center_of_mass)
from .resample_image import resample_image, resample_image_to_target
from .slice_image import slice_image
from .smooth_image import smooth_image
from .symmetrize_image import symmetrize_image
from .threshold_image import threshold_image
from .weingarten_image_curvature import weingarten_image_curvature



__all__ = ["nifti_to_ants"]

import numpy as np
import ants

def nifti_to_ants( nib_image ):
    """
    Converts a given Nifti image into an ANTsPy image

    Parameters
    ----------
        img: NiftiImage

    Returns
    -------
        ants_image: ANTsImage
    """
    ndim = nib_image.ndim

    if ndim < 3:
        print("Dimensionality is less than 3.")
        return None

    q_form = nib_image.get_qform()
    spacing = nib_image.header["pixdim"][1 : ndim + 1]

    origin = np.zeros((ndim))
    origin[:3] = q_form[:3, 3]

    direction = np.diag(np.ones(ndim))
    direction[:3, :3] = q_form[:3, :3] / spacing[:3]

    ants_img = ants.from_numpy(
        data = nib_image.get_data().astype( np.float ),
        origin = origin.tolist(),
        spacing = spacing.tolist(),
        direction = direction )
    
    return ants_img
